import aerosandbox as asb
import aerosandbox.numpy as np
import aerosandbox.tools.units as u
from typing import Dict, Union

"""
Applicable to:

* low-to-moderate performance (up to about 300 kt) light utility aircraft.
"""


def mass_landing_gear(
        gear_length: float,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
):
    """
    Calculates the mass of the landing gear.

    Args:

        gear_length: The length of the landing gear, in meters.

        design_mass_TOGW: The design takeoff gross weight of the aircraft, in kg.

        ultimate_load_factor: The ultimate load factor of the aircraft.

    Returns: The mass of the landing gear, in kg.
    """
    return (
            0.054 *
            (gear_length / u.inch) ** 0.501 *
            (design_mass_TOGW / u.lbm * ultimate_load_factor) ** 0.684
    ) * u.lbm



import aerosandbox as asb
import aerosandbox.numpy as np

AR = 2
taper = 0.1

af = asb.Airfoil("naca0010")
af.generate_polars(
    cache_filename="../cache/naca0010.json"
)

sweep_position_x_over_c = 1

wing = asb.Wing(
    symmetric=True,
    xsecs=[
        asb.WingXSec(
            xyz_le=[
                -sweep_position_x_over_c * taper ** -0.5,
                0,
                0
            ],
            chord=taper ** -0.5,
            airfoil=af
        ),
        asb.WingXSec(
            xyz_le=[
                -sweep_position_x_over_c * taper ** 0.5,
                AR,
                0
            ],
            chord=taper ** 0.5,
            airfoil=af
        )
    ]
).subdivide_sections(5)

airplane = asb.Airplane(
    wings=[wing]
)

alphas = np.linspace(0, 90, 500)

op_point = asb.OperatingPoint(
    atmosphere=asb.Atmosphere(altitude=0),
    velocity=10,
    alpha=alphas
)

aero = asb.AeroBuildup(
    airplane=airplane,
    op_point=op_point
).run()

if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots()
    plt.plot(
        alphas,
        aero["CL"],
        label="AeroBuildup"
    )

    from aerosandbox.tools.webplotdigitizer_reader import read_webplotdigitizer_csv

    polhamus_data = read_webplotdigitizer_csv(
        filename="data/wpd_datasets.csv"
    )

    polhamus_data_AR = polhamus_data[f'{AR:.1f}']

    plt.plot(
        polhamus_data_AR[:, 0],
        polhamus_data_AR[:, 1],
        ".",
        label="Experiment (Polhamus)"
    )
    plt.xlim(alphas.min(), alphas.max())
    plt.ylim(0, 1.5)

    p.set_ticks(10, 2, 0.1, 0.05)

    p.show_plot(
        f"AeroBuildup for Low-Aspect-Ratio, Swept Wings\n$AR={AR:.1f}$",
        r"Angle of Attack $\alpha$ [deg]",
        r"Lift Coefficient $C_L$ [-]"
    )



import aerosandbox.numpy as np
from aerosandbox.modeling.splines.hermite import linear_hermite_patch, cubic_hermite_patch


def sears_haack_drag(
        radius_max: float,
        length: float
) -> float:
    """
    Yields the idealized drag area (denoted CDA, or equivalently, D/q) of a Sears-Haack body.

    Assumes linearized supersonic (Prandtl-Glauert) flow.

    https://en.wikipedia.org/wiki/Sears%E2%80%93Haack_body

    Note that drag coefficient and drag area are independent of Mach number for this case (assuming linearized supersonic aero).

    Args:
        radius_max: The maximum radius of the Sears-Haack body.
        length: The length of the Sears-Haack body.

    Returns: The drag area (CDA, or D/q) of the body. To get the drag force, multiply by the dynamic pressure.

    """
    CDA = 9 * np.pi ** 2 * radius_max ** 2 / (2 * length ** 2)
    return CDA


def sears_haack_drag_from_volume(
        volume: float,
        length: float
) -> float:
    """
    See documentation for sears_haack_drag() in this same file.

    Identical, except takes volume as an input rather than max radius.

    Also returns a drag area (denoted CDA, or equivalently, D/q).
    """
    CDA = 128 * volume ** 2 / (np.pi * length ** 4)
    return CDA


def mach_crit_Korn(
        CL,
        t_over_c,
        sweep=0,
        kappa_A=0.95
):
    """
        Wave drag_force coefficient prediction using the low-fidelity Korn Equation method;
    derived in "Configuration Aerodynamics" by W.H. Mason, Sect. 7.5.2, pg. 7-18

    Args:
        CL: Sectional lift coefficient
        t_over_c: thickness-to-chord ratio
        sweep: sweep angle, in degrees
        kappa_A: Airfoil technology factor (0.95 for supercritical section, 0.87 for NACA 6-series)

    Returns:

    """
    smooth_abs_CL = np.softmax(CL, -CL, hardness=10)

    M_dd = kappa_A / np.cosd(sweep) - t_over_c / np.cosd(sweep) ** 2 - smooth_abs_CL / (10 * np.cosd(sweep) ** 3)
    M_crit = M_dd - (0.1 / 80) ** (1 / 3)
    return M_crit


def approximate_CD_wave(
        mach,
        mach_crit,
        CD_wave_at_fully_supersonic,
):
    """
    An approximate relation for computing transonic wave drag, based on an object's Mach number.

    Considered reasonably valid from Mach 0 up to around Mach 2 or 3-ish.

    Methodology is a combination of:

        * The methodology described in Raymer, "Aircraft Design: A Conceptual Approach", Section 12.5.10 Transonic Parasite Drag (pg. 449 in Ed. 2)

        and

        * The methodology described in W.H. Mason's Configuration Aerodynamics, Chapter 7. Transonic Aerodynamics of Airfoils and Wings.

    Args:

        mach: Mach number at the operating point to be evaluated

        mach_crit: Critical mach number, a function of the body geometry

        CD_wave_at_fully_supersonic: The wave drag coefficient of the body at the speed that it first goes (
        effectively) fully supersonic.

            Here, that is taken to mean at the Mach 1.2 case.

            This value should probably be derived using something similar to a Sears-Haack relation for the body in
            question, with a markup depending on geometry smoothness.

            The CD_wave predicted by this function will match this value exactly at M=1.2 and M=1.05.

            The peak CD_wave that is predicted is ~1.23 * this value, which occurs at M=1.10.

            In the high-Mach limit, this function asymptotes at 0.80 * this value, as empirically stated by Raymer.
            However, this model is only approximate and is likely not valid for high-supersonic flows.

    Returns: The approximate wave drag coefficient at the specified Mach number.

        The reference area is whatever the reference area used in the `CD_wave_at_fully_supersonic` parameter is.

    """
    mach_crit_max = 1 - (0.1 / 80) ** (1 / 3)

    mach_crit = -np.softmax(
        -mach_crit,
        -mach_crit_max,
        hardness=50
    )

    ### The following approximate relation is derived in W.H. Mason, "Configuration Aerodynamics", Chapter 7. Transonic Aerodynamics of Airfoils and Wings.
    ### Equation 7-8 on Page 7-19.
    ### This is in turn based on Lock's proposed empirically-derived shape of the drag rise, from Hilton, W.F., High Speed Aerodynamics, Longmans, Green & Co., London, 1952, pp. 47-49
    mach_dd = mach_crit + (0.1 / 80) ** (1 / 3)

    ### Model drag sections and cutoffs:
    return CD_wave_at_fully_supersonic * np.where(
        mach < mach_crit,
        0,
        np.where(
            mach < mach_dd,
            20 * (mach - mach_crit) ** 4,
            np.where(
                mach < 1.05,
                cubic_hermite_patch(
                    mach,
                    x_a=mach_dd,
                    x_b=1.05,
                    f_a=20 * (0.1 / 80) ** (4 / 3),
                    f_b=1,
                    dfdx_a=0.1,
                    dfdx_b=8
                ),
                np.where(
                    mach < 1.2,
                    cubic_hermite_patch(
                        mach,
                        x_a=1.05,
                        x_b=1.2,
                        f_a=1,
                        f_b=1,
                        dfdx_a=8,
                        dfdx_b=-4
                    ),
                    np.blend(
                        switch=4 * 2 * (mach - 1.2) / (1.2 - 0.8),
                        value_switch_high=0.8,
                        value_switch_low=1.2,
                    )
                    # 0.8 + 0.2 * np.exp(20 * (1.2 - mach))
                )
            )
        )
    )


if __name__ == '__main__':
    mc = 0.6
    drag = lambda mach: approximate_CD_wave(
        mach,
        mach_crit=mc,
        CD_wave_at_fully_supersonic=1,
    )

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots(1, 3, figsize=(10, 5))

    mach = np.linspace(0., 2, 10000)
    drag = drag(mach)
    ddragdm = np.gradient(drag, np.diff(mach)[0])
    dddragdm = np.gradient(ddragdm, np.diff(mach)[0])

    plt.sca(ax[0])
    plt.title("$C_D$")
    plt.ylabel("$C_{D, wave} / C_{D, wave, M=1.2}$")
    plt.plot(mach, drag)
    plt.ylim(-0.05, 1.5)

    # plt.ylim(-0.01, 0.05)

    plt.sca(ax[1])
    plt.title("$d(C_D)/d(M)$")
    plt.ylabel(r"$\frac{d(C_{D, wave})}{dM}$")
    plt.plot(mach, ddragdm)
    plt.ylim(-5, 15)

    plt.sca(ax[2])
    plt.title("$d^2(C_D)/d(M)^2$")
    plt.ylabel(r"$\frac{d^2(C_{D, wave})}{dM^2}$")
    plt.plot(mach, dddragdm)
    # plt.ylim(-5, 15)

    for a in ax:
        plt.sca(a)
        plt.xlim(0.6, 1.2)
        plt.xlabel("Mach [-]")

    p.show_plot()



from aerosandbox.modeling.fitting import FittedModel
from aerosandbox.modeling.interpolation import InterpolatedModel
from aerosandbox.modeling.interpolation_unstructured import UnstructuredInterpolatedModel



import aerosandbox.numpy as np
from aerosandbox.atmosphere.thermodynamics.gas import universal_gas_constant


def mass_flow_rate(
        mach,
        area,
        total_pressure,
        total_temperature,
        molecular_mass=28.9644e-3,
        gamma=1.4,
):
    specific_gas_constant = universal_gas_constant / molecular_mass
    return (
            (area * total_pressure) * (gamma / specific_gas_constant / total_temperature) ** 0.5
            * mach * (1 + (gamma - 1) / 2 * mach ** 2) ** (- (gamma + 1) / (2 * (gamma - 1)))
    )



import matplotlib.pyplot as plt


def hline(
        y,
        linestyle="--",
        color="k",
        text: str = None,
        text_xloc=0.5,
        text_ha="center",
        text_va="bottom",
        text_kwargs=None,
        **kwargs
):  # TODO docs
    if text_kwargs is None:
        text_kwargs = {}
    ax = plt.gca()
    xlim = ax.get_xlim()
    plt.axhline(y=y, ls=linestyle, color=color, **kwargs)
    if text is not None:
        plt.text(
            x=text_xloc * xlim[1] + (1 - text_xloc) * xlim[0],
            y=y,
            s=text,
            color=color,
            horizontalalignment=text_ha,
            verticalalignment=text_va,
            **text_kwargs
        )


def vline(
        x,
        linestyle="--",
        color="k",
        text: str = None,
        text_yloc=0.5,
        text_ha="right",
        text_va="center",
        text_kwargs=None,
        **kwargs
):  # TODO docs
    if text_kwargs is None:
        text_kwargs = {}
    ax = plt.gca()
    ylim = ax.get_ylim()
    plt.axvline(x=x, ls=linestyle, color=color, **kwargs)
    if text is not None:
        plt.text(
            x=x,
            y=text_yloc * ylim[1] + (1 - text_yloc) * ylim[0],
            s=text,
            color=color,
            horizontalalignment=text_ha,
            verticalalignment=text_va,
            rotation=90,
            **text_kwargs
        )



from datetime import datetime

import numpy as np
from matplotlib.dates import date2num


def ensure_float(value):
    """Make sure datetime values are properly converted to floats."""
    try:
        # the last 3 boolean checks are for arrays with datetime64 and with
        # a timezone, see these SO posts:
        # https://stackoverflow.com/q/60714568/4549682
        # https://stackoverflow.com/q/23063362/4549682
        # somewhere, the datetime64 with timezone is getting converted to 'O' dtype
        if (
                isinstance(value, datetime)
                or isinstance(value, np.datetime64)
                or np.issubdtype(value.dtype, np.datetime64)
                or str(value.dtype).startswith("datetime64")
                or value.dtype == "O"
        ):
            return date2num(value)
        else:  # another numpy dtype like float64
            return value
    except AttributeError:  # possibly int or other float/int dtype
        return value


# From https://www.geeksforgeeks.org/maximum-bipartite-matching/
class GFG:
    def __init__(self, graph):

        # residual graph
        self.graph = graph
        self.ppl = len(graph)
        self.jobs = len(graph[0])

    # A DFS based recursive function that returns true if a matching for vertex
    # u is possible
    def bpm(self, u, matchR, seen):

        # Try every job one by one
        for v in range(self.jobs):

            # If applicant u is interested
            # in job v and v is not seen
            if self.graph[u][v] and not seen[v]:

                # Mark v as visited
                seen[v] = True

                # If job 'v' is not assigned to an applicant OR previously
                # assigned applicant for job v (which is matchR[v]) has an
                # alternate job available. Since v is marked as visited in the
                # above line, matchR[v] in the following recursive call will not
                # get job 'v' again

                if matchR[v] == -1 or self.bpm(matchR[v], matchR, seen):
                    matchR[v] = u
                    return True
        return False

    # Returns maximum number of matching
    def maxBPM(self):
        # An array to keep track of the applicants assigned to jobs. The value
        # of matchR[i] is the applicant number assigned to job i, the value -1
        # indicates nobody is assigned.
        matchR = [-1] * self.jobs

        # Count of jobs assigned to applicants
        result = 0
        for i in range(self.ppl):

            # Mark all jobs as not seen for next applicant.
            seen = [False] * self.jobs

            # Find if the applicant 'u' can get a job
            if self.bpm(i, matchR, seen):
                result += 1
        return result, matchR


def maximum_bipartite_matching(graph: np.ndarray) -> np.ndarray:
    """Finds the maximum bipartite matching of a graph

    Parameters
    ----------
    graph : np.ndarray
        The graph, represented as a boolean matrix

    Returns
    -------
    order : np.ndarray
        The order in which to traverse the graph to visit a maximum of nodes
    """
    g = GFG(graph)
    _, order = g.maxBPM()
    return np.asarray(order)


def always_iterable(obj, base_type=(str, bytes)):
    """If *obj* is iterable, return an iterator over its items::

        >>> obj = (1, 2, 3)
        >>> list(always_iterable(obj))
        [1, 2, 3]

    If *obj* is not iterable, return a one-item iterable containing *obj*::

        >>> obj = 1
        >>> list(always_iterable(obj))
        [1]

    If *obj* is ``None``, return an empty iterable:

        >>> obj = None
        >>> list(always_iterable(None))
        []

    By default, binary and text strings are not considered iterable::

        >>> obj = 'foo'
        >>> list(always_iterable(obj))
        ['foo']

    If *base_type* is set, objects for which ``isinstance(obj, base_type)``
    returns ``True`` won't be considered iterable.

        >>> obj = {'a': 1}
        >>> list(always_iterable(obj))  # Iterate over the dict's keys
        ['a']
        >>> list(always_iterable(obj, base_type=dict))  # Treat dicts as a unit
        [{'a': 1}]

    Set *base_type* to ``None`` to avoid any special handling and treat objects
    Python considers iterable as iterable:

        >>> obj = 'foo'
        >>> list(always_iterable(obj, base_type=None))
        ['f', 'o', 'o']
    """
    if obj is None:
        return iter(())

    if (base_type is not None) and isinstance(obj, base_type):
        return iter((obj,))

    try:
        return iter(obj)
    except TypeError:
        return iter((obj,))



import aerosandbox.numpy as np
import casadi as cas
import pytest


def test_where_numpy():
    a = np.ones(4)
    b = 2 * np.ones(4)

    c = np.where(
        np.array([True, False, True, False]),
        a,
        b
    )

    assert np.all(
        c == np.array([1, 2, 1, 2])
    )


def test_where_casadi():
    a = cas.DM(np.ones(4))
    b = 2 * cas.DM(np.ones(4))

    c = np.where(
        cas.DM([1, 0, 1, 0]),
        a,
        b
    )

    assert np.all(
        c == cas.DM([1, 2, 1, 2])
    )


# def test_if_else_mixed(): # TODO write this


if __name__ == '__main__':
    pytest.main()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

from aerosandbox.library.aerodynamics import transonic


def test_simple_scalar_optimization():
    opti = asb.Opti()
    mach = opti.variable(init_guess=0.8, )
    CD_induced = 0.1 / mach
    CD_wave = transonic.approximate_CD_wave(
        mach=mach,
        mach_crit=0.7,
        CD_wave_at_fully_supersonic=1,
    )
    CD = CD_induced + CD_wave
    opti.minimize(CD)
    sol = opti.solve()


if __name__ == '__main__':
    pytest.main()



from typing import Union, Dict
import aerosandbox.numpy as np
from aerosandbox.modeling.surrogate_model import SurrogateModel


class InterpolatedModel(SurrogateModel):
    """
    A model that is interpolated to structured (i.e., gridded) N-dimensional data. Maps from R^N -> R^1.

    You can evaluate this model at a given point by calling it just like a function, e.g.:

    >>> y = my_interpolated_model(x)

    The input to the model (`x` in the example above) is of the type:
        * in the general N-dimensional case, a dictionary where: keys are variable names and values are float/array
        * in the case of a 1-dimensional input (R^1 -> R^1), it can optionally just be a float/array.
    If you're not sure what the input type of `my_interpolated_model` should be, just do:

    >>> print(my_interpolated_model) # Displays the valid input type to the model

    The output of the model (`y` in the example above) is always a float or array.

    See the docstring __init__ method of InterpolatedModel for more details of how to instantiate and use InterpolatedModel.

    One might have expected a interpolated model to be a literal Python function rather than a Python class - the
    benefit of having InterpolatedModel as a class rather than a function is that you can easily save (pickle) classes
    including data (e.g. parameters, x_data, y_data), but you can't do that with functions. And, because the
    InterpolatedModel class has a __call__ method, you can basically still just think of it like a function.

    """

    def __init__(self,
                 x_data_coordinates: Union[np.ndarray, Dict[str, np.ndarray]],
                 y_data_structured: np.ndarray,
                 method: str = "bspline",
                 fill_value=np.nan,  # Default behavior: return NaN for all inputs outside data range.
                 ):
        """
        Create the interpolator. Note that data must be structured (i.e., gridded on a hypercube) for general
        N-dimensional interpolation.

        Args:
            x_data_coordinates: The coordinates of each axis of the cube; essentially, the independent variable(s):

                * For the general N-dimensional case, this should be a dictionary where the keys are axis names [str]
                and the values are 1D arrays.

                * For the 1D case, you can optionally alternatively supply this as a single 1D array.

            Usage example for how you might generate this data, along with `y_data_structured`:

            >>> x1 = np.linspace(0, 5, 11)
            >>> x2 = np.linspace(0, 10, 21)
            >>> X1, X2 = np.meshgrid(x1, x2, indexing="ij")
            >>>
            >>> x_data_coordinates = {
            >>>     "x1": x1, # 1D ndarray of length 11
            >>>     "x2": x2, # 1D ndarray of length 21
            >>> }
            >>> y_data_structured = function_to_approximate(X1, X2) # 2D ndarray of shape (11, 21)

            y_data_structured: The dependent variable, expressed as a structured data "cube":

                * For the general N-dimensional case, this should be a single N-dimensional array with axis lengths
                corresponding to the inputs in `x_data_coordinates`. In the 1-dimensional case, this naturally
                reduces down to a single 1D ndarray.

                See usage example along with `x_data_coordinates` above.

            method: The method of interpolation to perform. Options:

                * "bspline" (Note: differentiable and suitable for optimization - made of piecewise-cubics. For other
                applications, other interpolators may be faster. Not monotonicity-preserving - may overshoot. Watch
                out for Runge's phenomenon; on that note, if your data is noisy, consider smoothing it first.)

                * "linear" (Note: differentiable, but not suitable for use in optimization w/o subgradient treatment due
                to C1-discontinuity)

                * "nearest" (Note: NOT differentiable, don't use in optimization. Fast.)

            fill_value: Gives the value that the interpolator should return for points outside of the interpolation
            domain. The interpolation domain is defined as the hypercube bounded by the coordinates specified in
            `x_data_coordinates`. If fill_value is None, then the interpolator will attempt to extrapolate if the interpolation method allows.

        """
        try:
            x_data_coordinates_values = x_data_coordinates.values()
        except AttributeError:  # If x_data_coordinates is not a dict
            x_data_coordinates_values = tuple([x_data_coordinates])

        ### Validate inputs
        for coordinates in x_data_coordinates_values:
            if len(coordinates.shape) != 1:
                raise ValueError("""
                    `x_data_coordinates` must be either: 
                        * In the general N-dimensional case, a dict where values are 1D ndarrays defining the coordinates of each axis.
                        * In the 1D case, can also be a 1D ndarray.
                    """)
        implied_y_data_shape = tuple(len(coordinates) for coordinates in x_data_coordinates_values)
        if not y_data_structured.shape == implied_y_data_shape:
            raise ValueError(f"""
            The shape of `y_data_structured` should be {implied_y_data_shape}
            """)

        ### Store data
        self.x_data_coordinates = x_data_coordinates
        self.x_data_coordinates_values = x_data_coordinates_values
        self.y_data_structured = y_data_structured
        self.method = method
        self.fill_value = fill_value

        ### Create unstructured versions of the data for plotting, etc.
        x_data = x_data_coordinates
        if isinstance(x_data, dict):
            x_data_values = np.meshgrid(*x_data_coordinates_values, indexing="ij")
            x_data = {
                k: v.reshape(-1)
                for k, v in zip(x_data_coordinates.keys(), x_data_values)
            }
        self.x_data = x_data
        self.y_data = np.ravel(y_data_structured, order="F")

    def __call__(self, x):
        if isinstance(self.x_data_coordinates, dict):
            def get_shape(value):
                if np.is_casadi_type(value, recursive=False):
                    if value.shape[1] == 1:
                        return (np.length(value),)

                try:
                    return value.shape
                except AttributeError:
                    return tuple()

            shape = np.broadcast_shapes(
                *[get_shape(v) for v in x.values()]
            )
            shape_for_reshaping = (int(np.product(shape)),)

            def reshape(value):
                try:
                    return np.reshape(value, shape_for_reshaping)
                except ValueError:
                    if isinstance(value, int) or isinstance(value, float) or value.shape == tuple() or np.product(
                            value.shape) == 1:
                        return value * np.ones(shape_for_reshaping)
                raise ValueError("Could not reshape value of one of the inputs!")

            x = np.stack(tuple(
                reshape(x[k])
                for k, v in self.x_data_coordinates.items()
            ), axis=1)

        output = np.interpn(
            points=self.x_data_coordinates_values,
            values=self.y_data_structured,
            xi=x,
            method=self.method,
            bounds_error=False,  # Can't be set true if general MX-type inputs are to be expected.
            fill_value=self.fill_value
        )
        try:
            return np.reshape(output, shape)
        except UnboundLocalError:
            return output



import numpy as _onp
import casadi as _cas
from aerosandbox.numpy.determine_type import is_casadi_type


def where(
        condition,
        value_if_true,
        value_if_false,
):
    """
    Return elements chosen from x or y depending on condition.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.where.html
    """
    if not is_casadi_type([condition, value_if_true, value_if_false], recursive=True):
        return _onp.where(
            condition,
            value_if_true,
            value_if_false
        )
    else:
        return _cas.if_else(
            condition,
            value_if_true,
            value_if_false
        )


def maximum(
        x1,
        x2,
):
    """
    Element-wise maximum of two arrays.

    Note: not differentiable at the crossover point, will cause issues if you try to optimize across a crossover.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.maximum.html
    """
    if not is_casadi_type([x1, x2], recursive=True):
        return _onp.maximum(x1, x2)
    else:
        return where(
            x1 >= x2,
            x1,
            x2,
        )


def minimum(
        x1,
        x2,
):
    """
    Element-wise minimum of two arrays.

    Note: not differentiable at the crossover point, will cause issues if you try to optimize across a crossover.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.minimum.html
    """
    if not is_casadi_type([x1, x2], recursive=True):
        return _onp.minimum(x1, x2)
    else:
        return where(
            x1 <= x2,
            x1,
            x2,
        )



from aerosandbox.structures.legacy.beams import *
import matplotlib.pyplot as plt
import seaborn as sns

sns.set(font_scale=1)

opti = cas.Opti()
beam = TubeBeam1(
    opti=opti,
    length=34 / 2,
    max_allowable_stress=570e6 / 1.75,
    bending=True,
    torsion=False
)
lift_force = 9.81 * 103.873
beam.add_uniform_load(force=lift_force / 2)
beam.setup()

# Tip deflection constraint
opti.subject_to([
    beam.u[-1] < 2,  # Source: http://web.mit.edu/drela/Public/web/hpa/hpa_structure.pdf
    beam.u[-1] > -2  # Source: http://web.mit.edu/drela/Public/web/hpa/hpa_structure.pdf
])

objective = beam.mass / 5

penalty = (
        cas.sum1((cas.diff(cas.diff(beam.nominal_diameter)) / 0.01) ** 2) / beam.n  # soft stabilizer
)

opti.minimize(objective + penalty)

p_opts = {}
s_opts = {}
s_opts["max_iter"] = 1e6  # If you need to interrupt, just use ctrl+c
s_opts["mu_strategy"] = "adaptive"
opti.solver('ipopt', p_opts, s_opts)

try:
    sol = opti.solve()
except:
    print("Warning: Failed!")
    sol = opti.debug

import copy

beam_sol = copy.deepcopy(beam).substitute_solution(sol)

print(f"Beam mass: {beam_sol.mass} kg")
print(f"Wing spar mass: {2 * beam_sol.mass} kg (Wing spar consists of two of these beams)")
beam_sol.draw_bending(show=False, for_print=True)

plt.savefig("validation_daedalus_plot.pgf")
plt.savefig("validation_daedalus_plot.pdf")
plt.show()



import aerosandbox.numpy as np
from aerosandbox.weights.mass_properties import MassProperties

"""
Most of these relations are taken from:
https://en.wikipedia.org/wiki/List_of_moments_of_inertia
"""


def mass_properties_from_radius_of_gyration(
        mass: float,
        x_cg: float = 0,
        y_cg: float = 0,
        z_cg: float = 0,
        radius_of_gyration_x: float = 0,
        radius_of_gyration_y: float = 0,
        radius_of_gyration_z: float = 0,
) -> MassProperties:
    """
    Returns the mass properties of an object, given its radius of gyration.

    It's assumed that the principle axes of the inertia tensor are aligned with the coordinate axes.

    This is a shorthand convenience function for common usage of the MassProperties constructor. For more detailed
    use, use the MassProperties object directly.

    Args:
        mass: Mass [kg]
        x_cg: x-position of the center of gravity
        y_cg: y-position of the center of gravity
        z_cg: z-position of the center of gravity
        radius_of_gyration_x: Radius of gyration along the x-axis, about the center of gravity [m]
        radius_of_gyration_y: Radius of gyration along the y-axis, about the center of gravity [m]
        radius_of_gyration_z: Radius of gyration along the z-axis, about the center of gravity [m]

    Returns: MassProperties object.

    """
    return MassProperties(
        mass=mass,
        x_cg=x_cg,
        y_cg=y_cg,
        z_cg=z_cg,
        Ixx=mass * radius_of_gyration_x ** 2,
        Iyy=mass * radius_of_gyration_y ** 2,
        Izz=mass * radius_of_gyration_z ** 2,
        Ixy=0,
        Iyz=0,
        Ixz=0,
    )


def mass_properties_of_ellipsoid(
        mass: float,
        radius_x: float,
        radius_y: float,
        radius_z: float,
) -> MassProperties:
    """
    Returns the mass properties of an ellipsoid centered on the origin.

    Args:
        mass: Mass [kg]
        radius_x: Radius along the x-axis [m]
        radius_y: Radius along the y-axis [m]
        radius_z: Radius along the z-axis [m]

    Returns: MassProperties object.

    """
    return MassProperties(
        mass=mass,
        x_cg=0,
        y_cg=0,
        z_cg=0,
        Ixx=0.2 * mass * (radius_y ** 2 + radius_z ** 2),
        Iyy=0.2 * mass * (radius_z ** 2 + radius_x ** 2),
        Izz=0.2 * mass * (radius_x ** 2 + radius_y ** 2),
        Ixy=0,
        Iyz=0,
        Ixz=0,
    )


def mass_properties_of_sphere(
        mass: float,
        radius: float,
) -> MassProperties:
    """
    Returns the mass properties of a sphere centered on the origin.

    Args:
        mass: Mass [kg]
        radius: Radius [m]

    Returns: MassProperties object.

    """
    return mass_properties_of_ellipsoid(
        mass=mass,
        radius_x=radius,
        radius_y=radius,
        radius_z=radius
    )


def mass_properties_of_rectangular_prism(
        mass: float,
        length_x: float,
        length_y: float,
        length_z: float,
) -> MassProperties:
    """
    Returns the mass properties of a rectangular prism centered on the origin.

    Args:
        mass: Mass [kg]
        length_x: Side length along the x-axis [m]
        length_y: Side length along the y-axis [m]
        length_z: Side length along the z-axis [m]

    Returns: MassProperties object.

    """
    return MassProperties(
        mass=mass,
        x_cg=0,
        y_cg=0,
        z_cg=0,
        Ixx=1 / 12 * mass * (length_y ** 2 + length_z ** 2),
        Iyy=1 / 12 * mass * (length_z ** 2 + length_x ** 2),
        Izz=1 / 12 * mass * (length_x ** 2 + length_y ** 2),
        Ixy=0,
        Iyz=0,
        Ixz=0,
    )


def mass_properties_of_cube(
        mass: float,
        side_length: float,
) -> MassProperties:
    """
    Returns the mass properties of a cube centered on the origin.

    Args:
        mass: Mass [kg]
        side_length: Side length of the cube [m]

    Returns: MassProperties object.

    """
    return mass_properties_of_rectangular_prism(
        mass=mass,
        length_x=side_length,
        length_y=side_length,
        length_z=side_length,
    )



import matplotlib.pyplot as plt
from matplotlib import ticker
from typing import Union
from .labellines import labelLines
import aerosandbox.numpy as np
from .threedim import ax_is_3d


def show_plot(
        title: str = None,
        xlabel: str = None,
        ylabel: str = None,
        dpi: float = None,
        savefig: str = None,
        tight_layout: bool = True,
        legend: bool = None,
        legend_inline: bool = False,
        legend_frame: bool = True,
        show: bool = True,
        pretty_grids: bool = True,
):
    """
    Finalize and show a plot.
    Args:
        title:
        xlabel:
        ylabel:
        tight_layout:
        legend:
        show:
        pretty_grids:

    Returns:

    """
    fig = plt.gcf()
    axes = fig.get_axes()

    if pretty_grids:
        for ax in axes:
            if not ax.get_label() == '<colorbar>':
                if not ax_is_3d(ax):
                    if any(line.get_visible() for line in ax.get_xgridlines()):
                        ax.grid(True, 'major', axis='x', linewidth=1.6)
                        ax.grid(True, 'minor', axis='x', linewidth=0.7)
                    if any(line.get_visible() for line in ax.get_ygridlines()):
                        ax.grid(True, 'major', axis='y', linewidth=1.6)
                        ax.grid(True, 'minor', axis='y', linewidth=0.7)
                else:
                    pass  # TODO
                    # if any(line.get_visible() for line in ax.get_xgridlines()):
                    #     ax.grid(True, 'major', axis='x', linewidth=1.6)
                    #     ax.grid(True, 'minor', axis='x', linewidth=0.7)
                    # if any(line.get_visible() for line in ax.get_ygridlines()):
                    #     ax.grid(True, 'major', axis='y', linewidth=1.6)
                    #     ax.grid(True, 'minor', axis='y', linewidth=0.7)
                    # if any(line.get_visible() for line in ax.get_zgridlines()):
                    #     ax.grid(True, 'major', axis='z', linewidth=1.6)
                    #     ax.grid(True, 'minor', axis='z', linewidth=0.7)

    ### Determine if a legend should be shown
    if legend is None:
        lines = plt.gca().lines
        if len(lines) <= 1:
            legend = False
        else:
            legend = False
            for line in lines:
                if line.get_label()[0] != "_":
                    legend = True
                    break

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if title is not None:
        if len(axes) <= 1:
            plt.title(title)
        else:
            plt.suptitle(title)
    if tight_layout:
        plt.tight_layout()
    if legend:
        if legend_inline:  # Display an inline (matplotlib-label-lines) legend instead
            labelLines(
                lines=plt.gca().get_lines(),
            )
        else:  # Display a traditional legend
            plt.legend(frameon=legend_frame)
    if dpi is not None:
        fig.set_dpi(dpi)
    if savefig is not None:
        plt.savefig(savefig)
    if show:
        plt.show()


def set_ticks(
        x_major: Union[float, int] = None,
        x_minor: Union[float, int] = None,
        y_major: Union[float, int] = None,
        y_minor: Union[float, int] = None,
        z_major: Union[float, int] = None,
        z_minor: Union[float, int] = None,
):
    ax = plt.gca()
    if x_major is not None:
        ax.xaxis.set_major_locator(ticker.MultipleLocator(base=x_major))
    if x_minor is not None:
        ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=x_minor))
    if y_major is not None:
        ax.yaxis.set_major_locator(ticker.MultipleLocator(base=y_major))
    if y_minor is not None:
        ax.yaxis.set_minor_locator(ticker.MultipleLocator(base=y_minor))
    if z_major is not None:
        ax.zaxis.set_major_locator(ticker.MultipleLocator(base=z_major))
    if z_minor is not None:
        ax.zaxis.set_minor_locator(ticker.MultipleLocator(base=z_minor))


def equal() -> None:
    """
    Sets all axes to be equal. Works for both 2d plots and 3d plots.

    Returns: None

    """
    ax = plt.gca()

    if not ax_is_3d(ax):
        ax.set_aspect("equal", adjustable='box')

    else:
        ax.set_box_aspect((1, 1, 1))

        xlim = ax.get_xlim3d()
        ylim = ax.get_ylim3d()
        zlim = ax.get_zlim3d()

        x_range = abs(xlim[1] - xlim[0])
        x_middle = np.mean(xlim)
        y_range = abs(ylim[1] - ylim[0])
        y_middle = np.mean(ylim)
        z_range = abs(zlim[1] - zlim[0])
        z_middle = np.mean(zlim)

        # The plot bounding box is a sphere in the sense of the infinity
        # norm, hence I call half the max range the plot radius.
        plot_radius = 0.5 * max([x_range, y_range, z_range])

        ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
        ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
        ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])



import numpy as np
from scipy import interpolate


class NaturalUnivariateSpline(interpolate.UnivariateSpline):
    """
    A Natural UnivariateSpline.

    Identical to a UnivariateSpline, except that extrapolation outside the data range is constrained to be linear.

    Based on: https://bobby.gramacy.com/surrogates/splines.html

    Which paraphrases [Hastie, Tibshirani & Friedman (2017)](https://hastie.su.domains/ElemStatLearn/), Chapters 5, 7, & 8.

    """

    def __init__(self,
                 x,
                 y,
                 w=None,
                 bbox=[None] * 2,
                 k=3,
                 s=None,
                 ext=0,
                 check_finite=False
                 ):
        super().__init__(
            x=x,
            y=y,
            w=w,
            bbox=bbox,
            k=k,
            s=s,
            ext=ext,
            check_finite=check_finite
        )
        self.x = x
        self.y = y
        self.w = w

    def __call__(self, x, nu=0, ext=None):
        xmin = np.min(self.x)
        xmax = np.max(self.x)

        cubic = super().__call__

        return np.where(
            x < xmin,
            cubic(xmin) + cubic(xmin, nu=1) * (x - xmin),
            np.where(
                x > xmax,
                cubic(xmax) + cubic(xmax, nu=1) * (x - xmax),
                cubic(x)
            )
        )


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    x = np.linspace(0, 10)
    y = np.sin(x)

    interpolator = NaturalUnivariateSpline(
        x,
        y,
        # s=0
    )

    x_plot = np.linspace(-5, 15)
    y_plot = interpolator(x_plot)

    fig, ax = plt.subplots()
    plt.plot(x, y, ".k", label="Data")
    plt.plot(x_plot, y_plot, "--", label="Interpolation")
    p.show_plot()



import aerosandbox as asb
import aerosandbox.numpy as np

import matplotlib.pyplot as plt
import aerosandbox.tools.pretty_plots as p

fuselage = asb.Fuselage(
    xsecs=[
        asb.FuselageXSec(
            xyz_c=[xi, 0, 0],
            radius=asb.Airfoil("naca0020").local_thickness(xi) / 2
        )
        for xi in np.cosspace(0, 1, 20)
    ],
)

fig, ax = plt.subplots(figsize=(7, 6))
V = np.linspace(10, 1000, 1001)
op_point = asb.OperatingPoint(
    velocity=V,
)

aero = asb.AeroBuildup(
    airplane=asb.Airplane(
        fuselages=[fuselage]
    ),
    op_point=op_point,
).run()

plt.plot(
    op_point.mach(),
    aero["CD"],
    label="Full Model"
)

aero = asb.AeroBuildup(
    airplane=asb.Airplane(
        fuselages=[fuselage]
    ),
    op_point=op_point,
    include_wave_drag=False
).run()

plt.plot(
    op_point.mach(),
    aero["CD"],
    zorder=1.9,
    label="Model without Wave Drag"
)
p.show_plot(
    "Transonic Fuselage Drag",
    "Mach [-]",
    "Drag Area $C_D \cdot A$ [m$^2$]"
)

print("%.4g" % aero["CD"][-1])



import numpy as np

from aerosandbox import AeroSandboxObject
from aerosandbox.geometry.common import *
from typing import List, Dict, Any, Union, Tuple, Optional, Callable
import copy


class Fuselage(AeroSandboxObject):
    """
    Definition for a Fuselage or other slender body (pod, fuel tank, etc.).

    Anatomy of a Fuselage:

        A fuselage consists chiefly of a collection of cross-sections, or "xsecs". A cross-section is a 2D "slice" of
        a fuselage. These can be accessed with `Fuselage.xsecs`, which gives a list of xsecs in the Fuselage. Each
        xsec is a FuselageXSec object, a class that is defined separately.

        You may also see references to fuselage "sections", which are different from cross-sections (xsecs)! Sections
        are the portions of the fuselage that are in between xsecs. In other words, a fuselage with N cross-sections
        (xsecs, FuselageXSec objects) will always have N-1 sections. Sections are never explicitly defined,
        since you can get all needed information by lofting from the adjacent cross-sections. For example,
        section 0 (the first one) is a loft between cross-sections 0 and 1.

        Fuselages are lofted linearly between cross-sections.

    """

    def __init__(self,
                 name: Optional[str] = "Untitled",
                 xsecs: List['FuselageXSec'] = None,
                 analysis_specific_options: Optional[Dict[type, Dict[str, Any]]] = None,
                 **kwargs,  # Only to allow for capturing of deprecated arguments, don't use this.
                 ):
        """
        Defines a new fuselage.

        Args:

            name: Name of the fuselage [optional]. It can help when debugging to give each fuselage a sensible name.

            xsecs: A list of fuselage cross-sections ("xsecs") in the form of FuselageXSec objects.

            analysis_specific_options: Analysis-specific options are additional constants or modeling assumptions
            that should be passed on to specific analyses and associated with this specific geometry object.

                This should be a dictionary where:

                    * Keys are specific analysis types (typically a subclass of asb.ExplicitAnalysis or
                    asb.ImplicitAnalysis), but if you decide to write your own analysis and want to make this key
                    something else (like a string), that's totally fine - it's just a unique identifier for the
                    specific analysis you're running.

                    * Values are a dictionary of key:value pairs, where:

                        * Keys are strings.

                        * Values are some value you want to assign.

                This is more easily demonstrated / understood with an example:

                >>> analysis_specific_options = {
                >>>     asb.AeroBuildup: dict(
                >>>         include_wave_drag=True,
                >>>     )
                >>> }
        """
        ### Set defaults
        if xsecs is None:
            xsecs: List['FuselageXSec'] = []
        if analysis_specific_options is None:
            analysis_specific_options = {}

        ### Initialize
        self.name = name
        self.xsecs = xsecs
        self.analysis_specific_options = analysis_specific_options

        ### Handle deprecated parameters
        if 'symmetric' in locals():
            raise Exception(
                "The `symmetric` argument for Fuselage objects is deprecated. Make your fuselages separate instead!")

        if 'xyz_le' in locals():
            import warnings
            warnings.warn(
                "The `xyz_le` input for Fuselage is DEPRECATED and will be removed in a future version. Use Fuselage().translate(xyz) instead.",
                stacklevel=2
            )
            self.xsecs = [
                xsec.translate(xyz_le)
                for xsec in self.xsecs
            ]

    def __repr__(self) -> str:
        n_xsecs = len(self.xsecs)
        return f"Fuselage '{self.name}' ({len(self.xsecs)} {'xsec' if n_xsecs == 1 else 'xsecs'})"

    def add_loft(self,
                 kind: str,
                 to_xsec: 'FuselageXSec',
                 from_xsec: 'FuselageXSec' = None,
                 n_points: int = 5,
                 spacing: Callable[[float, float, int], np.ndarray] = np.cosspace,
                 ) -> "Fuselage":
        raise NotImplementedError # Function under construction!
        ### Set defaults
        if from_xsec is None:
            if len(self.xsecs) == 0:
                from_xsec = FuselageXSec(
                    xyz_c=[0, 0, 0],
                    width=0,
                    height=0,
                    shape=2
                )
            else:
                from_xsec = self.xsecs[-1]

        ### Define a nondimensional coordinate
        t = spacing(0, 1, n_points)

        if kind == "linear":
            new_xsecs = [
                FuselageXSec(
                    xyz_c=from_xsec.xyz_c * (1 - ti) + to_xsec.xyz_c * ti,
                    width=from_xsec.width * (1 - ti) + to_xsec.width * ti,
                    height=from_xsec.height * (1 - ti) + to_xsec.height * ti,
                    shape=from_xsec.shape * (1 - ti) + to_xsec.shape * ti,
                    analysis_specific_options=from_xsec.analysis_specific_options,
                )
                for ti in t
            ]
        elif kind == "ellipsoid-nose":
            new_xsecs = [
                FuselageXSec(
                    xyz_c=from_xsec.xyz_c * (1 - ti) + to_xsec.xyz_c * ti,
                    width=from_xsec.width * (1 - ti) + to_xsec.width * ti,
                    height=from_xsec.height * (1 - ti) + to_xsec.height * ti,
                    shape=from_xsec.shape * (1 - ti) + to_xsec.shape * ti,
                    analysis_specific_options=from_xsec.analysis_specific_options,
                )
                for ti in t
            ]

        self.xsecs.extend(new_xsecs)

    def translate(self,
                  xyz: Union[np.ndarray, List[float]]
                  ) -> "Fuselage":
        """
        Translates the entire Fuselage by a certain amount.

        Args:
            xyz:

        Returns: self

        """
        new_fuse = copy.copy(self)
        new_fuse.xsecs = [
            xsec.translate(xyz)
            for xsec in new_fuse.xsecs
        ]
        return new_fuse

    def area_wetted(self) -> float:
        """
        Returns the wetted area of the fuselage.

        :return:
        """
        area = 0

        perimeters = [xsec.xsec_perimeter() for xsec in self.xsecs]

        for i in range(len(self.xsecs) - 1):
            x_separation = self.xsecs[i + 1].xyz_c[0] - self.xsecs[i].xyz_c[0]
            area += (perimeters[i] + perimeters[i + 1]) / 2 * x_separation

        return area

    def area_projected(self,
                       type: str = "XY",
                       ) -> float:
        """
        Returns the area of the fuselage as projected onto one of the principal planes.

        Args:
            type: A string, which determines which principal plane to use for projection. One of:

                * "XY", in which case the projected area is onto the XY plane (i.e., top-down)

                * "XZ", in which case the projected area is onto the XZ plane (i.e., side-view)

        Returns: The projected area.
        """
        area = 0
        for i in range(len(self.xsecs) - 1):
            x_separation = self.xsecs[i + 1].xyz_c[0] - self.xsecs[i].xyz_c[0]

            if type == "XY":
                width_a = self.xsecs[i].width
                width_b = self.xsecs[i + 1].width
                area += (width_a + width_b) / 2 * x_separation
            elif type == "XZ":
                height_a = self.xsecs[i].height
                height_b = self.xsecs[i + 1].height
                area += (height_a + height_b) / 2 * x_separation
            else:
                raise ValueError("Bad value of `type`!")

        return area

    def area_base(self) -> float:
        """
        Returns the area of the base (i.e. "trailing edge") of the fuselage. Useful for certain types of drag
        calculation.

        Returns:
        """
        return self.xsecs[-1].xsec_area()

    def fineness_ratio(
            self,
            assumed_shape="cylinder",
    ) -> float:
        """
        Approximates the fineness ratio using the volume and length. The fineness ratio of a fuselage is defined as:

            FR = length / max_diameter

        Args:

            assumed_shape: A string, which determines the assumed shape of the fuselage for the approximation. One of:

                * "cylinder", in which case the fuselage is assumed to have a cylindrical shape.

                * "sears-haack", in which case the fuselage is assumed to have Sears-Haack fuselage shape.

        Returns: An approximate value of the fuselage's fineness ratio.

        """
        if assumed_shape == "cylinder":
            return np.sqrt(
                self.length() ** 3 / self.volume() * np.pi / 4
            )
        elif assumed_shape == "sears-haack":
            length = self.length()

            r_max = np.sqrt(
                self.volume() / length / (3 * np.pi ** 2 / 16)
            )
            return length / r_max

    def length(self) -> float:
        """
        Returns the total front-to-back length of the fuselage. Measured as the difference between the x-coordinates
        of the leading and trailing cross-sections.
        :return:
        """
        return np.fabs(self.xsecs[-1].xyz_c[0] - self.xsecs[0].xyz_c[0])

    def volume(self,
               _sectional: bool = False
               ) -> Union[float, List[float]]:
        """
        Computes the volume of the Fuselage.

        Args:

            _sectional: A boolean. If False, returns the total volume. If True, returns a list of volumes for each of
            the `n-1` lofted sections (between the `n` fuselage cross-sections in fuselage.xsec).

        Returns:

            The computed volume.
        """
        xsec_areas = [
            xsec.xsec_area()
            for xsec in self.xsecs
        ]

        separations = [
            xsec_b.xyz_c[0] - xsec_a.xyz_c[0]
            for xsec_a, xsec_b in zip(
                self.xsecs[:-1],
                self.xsecs[1:]
            )
        ]

        sectional_volumes = [
            separation / 3 * (area_a + area_b + (area_a * area_b + 1e-100) ** 0.5)
            for area_a, area_b, separation in zip(
                xsec_areas[1:],
                xsec_areas[:-1],
                separations
            )
        ]

        volume = sum(sectional_volumes)

        if _sectional:
            return sectional_volumes
        else:
            return volume

    def x_centroid_projected(self,
                             type: str = "XY",
                             ) -> float:
        """
        Returns the x_g coordinate of the centroid of the planform area.

        Args:
            type: A string, which determines which principal plane to use for projection. One of:

                * "XY", in which case the projected area is onto the XY plane (i.e., top-down)

                * "XZ", in which case the projected area is onto the XZ plane (i.e., side-view)

        Returns: The x_g coordinate of the centroid.
        """

        total_x_area_product = 0
        total_area = 0

        for xsec_a, xsec_b in zip(self.xsecs, self.xsecs[1:]):
            x_a = xsec_a.xyz_c[0]
            x_b = xsec_b.xyz_c[0]

            if type == "XY":
                r_a = xsec_a.width / 2
                r_b = xsec_b.width / 2
            elif type == "XZ":
                r_a = xsec_a.height / 2
                r_b = xsec_b.height / 2
            else:
                raise ValueError("Bad value of `type`!")

            dx = x_b - x_a

            x_c = x_a + (r_a + 2 * r_b) / (3 * (r_a + r_b)) * dx
            area = (r_a + r_b) / 2 * dx

            total_area += area
            total_x_area_product += x_c * area

        x_centroid = total_x_area_product / total_area
        return x_centroid

    def mesh_body(self,
                  method="quad",
                  tangential_resolution: int = 36,
                  ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Meshes the fuselage as a solid (thickened) body.

        Uses the `(points, faces)` standard mesh format. For reference on this format, see the documentation in
        `aerosandbox.geometry.mesh_utilities`.

        Args:

            method: A string, which determines whether to mesh the fuselage as a series of quadrilaterals or triangles.

                * "quad" meshes the fuselage as a series of quadrilaterals.

                * "tri" meshes the fuselage as a series of triangles.

            tangential_resolution: An integer, which determines the number of points to use to mesh each cross-section.

        Returns: Standard unstructured mesh format: A tuple of`points` and `faces`, where:

            * `points` is a `n x 3` array of points, where `n` is the number of points in the mesh.

            * `faces` is a `m x 3` array of faces if `method` is "tri", or a `m x 4` array of faces if `method` is "quad".

                * Each row of `faces` is a list of indices into `points`, which specifies a face.

        """

        t = np.linspace(0, 2 * np.pi, tangential_resolution + 1)[:-1]

        points = np.concatenate([
            np.stack(
                xsec.get_3D_coordinates(theta=t),
                axis=1
            )
            for xsec in self.xsecs
        ],
            axis=0
        )

        faces = []

        num_i = len(self.xsecs)
        num_j = len(t)

        def index_of(iloc, jloc):
            return iloc * num_j + (jloc % num_j)

        def add_face(*indices):
            entry = list(indices)
            if method == "quad":
                faces.append(entry)
            elif method == "tri":
                faces.append([entry[0], entry[1], entry[3]])
                faces.append([entry[1], entry[2], entry[3]])

        for i in range(num_i - 1):
            for j in range(num_j):
                add_face(
                    index_of(i, j),
                    index_of(i, j + 1),
                    index_of(i + 1, j + 1),
                    index_of(i + 1, j),
                )

        faces = np.array(faces)

        return points, faces

    def mesh_line(self,
                  y_nondim: Union[float, List[float]] = 0.,
                  z_nondim: Union[float, List[float]] = 0.,
                  ) -> List[np.ndarray]:
        """
        Returns points along a line that goes through each of the FuselageXSec objects in this Fuselage.

        Args:

            y_nondim: The nondimensional (width-normalized) y-coordinate that the line should go through. Can either
            be a single value used at all cross-sections, or can be an iterable of values to be used at the
            respective cross-sections.

            z_nondim: The nondimensional (height-normalized) z-coordinate that the line should go through. Can either
            be a single value used at all cross-sections, or can be an iterable of values to be used at the
            respective cross-sections.

        Returns: A list of points, where each point is a 3-element array of the form `[x, y, z]`. Goes from the nose
        to the tail.

        """

        points_on_line: List[np.ndarray] = []

        try:
            if len(y_nondim) != len(self.xsecs):
                raise ValueError(
                    f"If `y_nondim` is an iterable, it should be the same length as `Fuselage.xsecs` ({len(self.xsecs)})."
                )
        except TypeError:
            pass

        try:
            if len(z_nondim) != len(self.xsecs):
                raise ValueError(
                    f"If `z_nondim` is an iterable, it should be the same length as `Fuselage.xsecs` ({len(self.xsecs)})."
                )
        except TypeError:
            pass

        for i, xsec in enumerate(self.xsecs):

            origin = xsec.xyz_c
            xg_local, yg_local, zg_local = xsec.compute_frame()

            try:
                xsec_y_nondim = y_nondim[i]
            except (TypeError, IndexError):
                xsec_y_nondim = y_nondim

            try:
                xsec_z_nondim = z_nondim[i]
            except (TypeError, IndexError):
                xsec_z_nondim = z_nondim

            xsec_point = origin + (
                    xsec_y_nondim * (xsec.width / 2) * yg_local +
                    xsec_z_nondim * (xsec.height / 2) * zg_local
            )
            points_on_line.append(xsec_point)

        return points_on_line

    def draw(self, *args, **kwargs):
        """
        An alias to the more general Airplane.draw() method. See there for documentation.

        Args:
            *args: Arguments to pass through to Airplane.draw()
            **kwargs: Keyword arguments to pass through to Airplane.draw()

        Returns: Same return as Airplane.draw()

        """
        from aerosandbox.geometry.airplane import Airplane
        return Airplane(fuselages=[self]).draw(*args, **kwargs)

    def draw_wireframe(self, *args, **kwargs):
        """
        An alias to the more general Airplane.draw_wireframe() method. See there for documentation.

        Args:
            *args: Arguments to pass through to Airplane.draw_wireframe()
            **kwargs: Keyword arguments to pass through to Airplane.draw_wireframe()

        Returns: Same return as Airplane.draw_wireframe()

        """
        from aerosandbox.geometry.airplane import Airplane
        return Airplane(fuselages=[self]).draw_wireframe(*args, **kwargs)

    def draw_three_view(self, *args, **kwargs):
        """
        An alias to the more general Airplane.draw_three_view() method. See there for documentation.

        Args:
            *args: Arguments to pass through to Airplane.draw_three_view()
            **kwargs: Keyword arguments to pass through to Airplane.draw_three_view()

        Returns: Same return as Airplane.draw_three_view()

        """
        from aerosandbox.geometry.airplane import Airplane
        return Airplane(fuselages=[self]).draw_three_view(*args, **kwargs)

    def subdivide_sections(self,
                           ratio: int,
                           spacing_function: Callable[[float, float, float], np.ndarray] = np.linspace
                           ) -> "Fuselage":
        """
        Generates a new Fuselage that subdivides the existing sections of this Fuselage into several smaller ones. Splits
        each section into N=`ratio` smaller subsections by inserting new cross-sections (xsecs) as needed.

        This can allow for finer aerodynamic resolution of sectional properties in certain analyses.

        Args:

            ratio: The number of new sections to split each old section into.

            spacing_function: A function that takes in three arguments: the start, end, and number of points to generate.

                The default is `np.linspace`, which generates a linearly-spaced array of points.

                Other options include `np.cosspace`, which generates a cosine-spaced array of points.

        Returns: A new Fuselage object with subdivided sections.

        """
        if not (ratio >= 2 and isinstance(ratio, int)):
            raise ValueError("`ratio` must be an integer greater than or equal to 2.")

        new_xsecs = []
        length_fractions_along_section = spacing_function(0, 1, ratio + 1)[:-1]

        for xsec_a, xsec_b in zip(self.xsecs[:-1], self.xsecs[1:]):
            for s in length_fractions_along_section:
                a_weight = 1 - s
                b_weight = s

                new_xsecs.append(
                    FuselageXSec(
                        xyz_c=xsec_a.xyz_c * a_weight + xsec_b.xyz_c * b_weight,
                        width=xsec_a.width * a_weight + xsec_b.width * b_weight,
                        height=xsec_a.height * a_weight + xsec_b.height * b_weight,
                        shape=xsec_a.shape * a_weight + xsec_b.shape * b_weight,
                        analysis_specific_options=xsec_a.analysis_specific_options,
                    )
                )

        new_xsecs.append(self.xsecs[-1])

        return Fuselage(
            name=self.name,
            xsecs=new_xsecs,
            analysis_specific_options=self.analysis_specific_options
        )

    def _compute_frame_of_FuselageXSec(self, index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes the local frame of a FuselageXSec, given the index of the FuselageXSec in the Fuselage.xsecs list.

        Args:
            index: The index of the FuselageXSec in the Fuselage.xsecs list.

        Returns:  A tuple:
            xg_local: The x-axis of the local coordinate frame, in aircraft geometry axes.
            yg_local: The y-axis of the local coordinate frame, in aircraft geometry axes.
            zg_local: The z-axis of the local coordinate frame, in aircraft geometry axes.
        """
        import warnings
        warnings.warn(
            "Fuselage._compute_frame_of_FuselageXSec() is deprecated. "
            "Use FuselageXSec.compute_frame() instead.",
            DeprecationWarning
        )

        return self.xsecs[index].compute_frame()


class FuselageXSec(AeroSandboxObject):
    """
    Definition for a fuselage cross-section ("X-section").
    """

    def __init__(self,
                 xyz_c: Union[np.ndarray, List[float]] = None,
                 xyz_normal: Union[np.ndarray, List[float]] = None,
                 radius: float = None,
                 width: float = None,
                 height: float = None,
                 shape: float = 2.,
                 analysis_specific_options: Optional[Dict[type, Dict[str, Any]]] = None,
                 ):
        """
        Defines a new Fuselage cross-section.

        Fuselage cross-sections are essentially a sketch on a 2D plane.

            * This plane is defined by a center point (`xyz_c`) and a normal vector (`xyz_normal`).

            * The cross-section is a superellipse shape, which is a generalization of a circle and a square.

                It is mathematically defined by three parameters, using `y` and `z` as the two axes:

                    abs(y / width) ^ shape + abs(z / height) ^ shape = 1

                See also: https://en.wikipedia.org/wiki/Superellipse

                There are some notable special cases:

                    * A circle is a special case of a superellipse, where `shape = 2`.

                    * A square is a special case of a superellipse, where `shape = Inf` (in practice, set this to some
                    high value like 1000).

                    * A diamond is a special case of a superellipse, where `shape = 1`.

        Must specify either `radius` or both `width` and `height`. Cannot specify both.

        Args:

            xyz_c: An array-like that represents the xyz-coordinates of the center of this fuselage cross-section,
            in geometry axes.

            xyz_normal: An array-like that represents the xyz-coordinates of the normal vector of this fuselage
            cross-section, in geometry axes.

            radius: Radius of the fuselage cross-section.

            width: Width of the fuselage cross-section.

            height: Height of the fuselage cross-section.

            shape: A parameter that determines what shape the cross-section is. Should be in the range 1 < shape < infinity.

                In short, here's how to interpret this value:

                    * shape=2 is a circle.

                    * shape=1 is a diamond shape.

                    * A high value of, say, 10, will get you a square-ish shape.

                To be more precise:

                    * If the `shape` parameter is `s`, then the corresponding shape is the same as a level-set of a L^s norm in R^2.

                    * Defined another way, if the `shape` parameter is `s`, then the shape is the solution to the equation:

                        * x^s + y^s = 1 in the first quadrant (x>0, y>0); then mirrored for all four quadrants.

            analysis_specific_options: Analysis-specific options are additional constants or modeling assumptions
            that should be passed on to specific analyses and associated with this specific geometry object.

                This should be a dictionary where:

                    * Keys are specific analysis types (typically a subclass of asb.ExplicitAnalysis or
                    asb.ImplicitAnalysis), but if you decide to write your own analysis and want to make this key
                    something else (like a string), that's totally fine - it's just a unique identifier for the
                    specific analysis you're running.

                    * Values are a dictionary of key:value pairs, where:

                        * Keys are strings.

                        * Values are some value you want to assign.

                This is more easily demonstrated / understood with an example:

                >>> analysis_specific_options = {
                >>>     asb.AeroBuildup: dict(
                >>>         include_wave_drag=True,
                >>>     )
                >>> }
        """
        ### Set defaults
        if xyz_c is None:
            xyz_c = np.array([0., 0., 0.])
        if xyz_normal is None:
            xyz_normal = np.array([1., 0., 0.])  # points backwards
        if analysis_specific_options is None:
            analysis_specific_options = {}

        ### Set width and height
        radius_specified = (radius is not None)
        width_height_specified = [
            (width is not None),
            (height is not None)
        ]

        if radius_specified:
            if any(width_height_specified):
                raise ValueError(
                    "Cannot specify both `radius` and (`width`, `height`) parameters - must be one or the other."
                )

            self.width = 2 * radius
            self.height = 2 * radius

        else:
            if not all(width_height_specified):
                raise ValueError(
                    "Must specify either `radius` or both (`width`, `height`) parameters."
                )
            self.width = width
            self.height = height

        ### Initialize
        self.xyz_c = np.array(xyz_c)
        self.xyz_normal = np.array(xyz_normal)
        self.shape = shape
        self.analysis_specific_options = analysis_specific_options

    def __repr__(self) -> str:
        return f"FuselageXSec (xyz_c: {self.xyz_c}, width: {self.width}, height: {self.height}, shape: {self.shape})"

    def xsec_area(self):
        """
        Computes the FuselageXSec's cross-sectional (xsec) area.

        The computation method is a closed-form approximation for the area of a superellipse. The exact equation for
        the area of a superellipse with shape parameter `s` is:

            area = width * height * (gamma(1 + 1/n))^2 / gamma(1 + 2/n)

        where gamma() is the gamma function. The gamma function is (relatively) computationally expensive to evaluate
        and differentiate, so we replace this area calculation with a closed-form approximation (with essentially no
        loss in accuracy):

            area = width * height / (s^-1.8717618013591173 + 1)

        This approximation has the following properties:

            * It is numerically exact for the case of s = 1 (a diamond)

            * It is numerically exact for the case of s = 2 (a circle)

            * It is correct in the asymptotic limit where s -> infinity (a square)

            * In the range of sensible s values (1 < s < infinity), its error is less than 0.6%.

            * It always produces a positive area for any physically-meaningful value of s (s > 0). In the range of s
            values where s is physically-meaningful but not in a sensible range (0 < s < 1), this equation will
            over-predict area.

        The value of the constant seen in this expression (1.872...) is given by log(4/pi - 1) / log(2), and it is
        chosen as such so that the expression is exactly correct in the s=2 (circle) case.

        Returns:

        """
        area = self.width * self.height / (self.shape ** -1.8717618013591173 + 1)

        return area

    def xsec_perimeter(self):
        """
        Computes the FuselageXSec's perimeter. ("Circumference" in the case of a circular cross-section.)

        The computation method is a closed-form approximation for the perimeter of a superellipse. The exact equation
        for the perimeter of a superellipse is quite long and is not repeated here for brevity; a Google search will
        bring it up. More importantly, this exact equation can only be represented as an infinite sum - not
        particularly useful for fast computation.

        We replace this exact equation with the following closed-form approximation obtained from symbolic regression:

            Imagine a superellipse centered on the origin of a 2D plane. Now, imagine that the superellipse is
            stretched such that the first quadrant (e.g., x>0, y>0) goes from (1, 0) to (0, h). Assume it has shape
            parameter s (where, as a reminder, s=1 is a diamond, s=2 is a circle, s=Inf is a square).

            Then, the perimeter of that single quadrant is:

            h + (((((s-0.88487077) * h + 0.2588574 / h) ^ exp(s / -0.90069205)) + h) + 0.09919785) ^ (-1.4812293 / s)

            See `AeroSandbox/studies/SuperellipseProperties` for details about how this was obtained.

        We can extrapolate from here to the general case of a superellipse, as shown in the code below.

        This approximation has the following properties:

            * For the s=1 case (diamond), the error is +0.2%.

            * For the s=2 case (circle), the error is -0.1%.

            * In the s -> infinity limit (square), the error is +0.1%.

        Returns:

        """
        try:
            if self.width == 0:
                return 2 * self.height
            elif self.height == 0:
                return 2 * self.width
        except RuntimeError:  # Will error if width and height are optimization variables, as truthiness is indeterminate
            pass

        s = self.shape
        h = np.maximum(
            (self.width + 1e-16) / (self.height + 1e-16),
            (self.height + 1e-16) / (self.width + 1e-16)
        )
        nondim_quadrant_perimeter = (
                h + (((((s - 0.88487077) * h + 0.2588574 / h) ** np.exp(s / -0.90069205)) + h) + 0.09919785) ** (
                -1.4812293 / s)
        )
        perimeter = 2 * nondim_quadrant_perimeter * np.minimum(self.width, self.height)

        return np.where(
            self.width == 0,
            2 * self.height,
            np.where(
                self.height == 0,
                2 * self.width,
                perimeter
            )
        )

    def compute_frame(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes the local coordinate frame of the FuselageXSec, in aircraft geometry axes.

        xg_local is aligned with the FuselageXSec's normal vector.

        zg_local is roughly aligned with the z-axis of the aircraft geometry axes, but projected onto the FuselageXSec's plane.

        yg_local is the cross product of zg_local and xg_local.

        Returns: A tuple:
            xg_local: The x-axis of the local coordinate frame, in aircraft geometry axes.
            yg_local: The y-axis of the local coordinate frame, in aircraft geometry axes.
            zg_local: The z-axis of the local coordinate frame, in aircraft geometry axes.

        """
        xyz_normal = self.xyz_normal / np.linalg.norm(self.xyz_normal)

        xg_local = xyz_normal

        zg_local = np.array([0, 0, 1])
        zg_local = zg_local - np.dot(zg_local, xg_local) * xg_local

        yg_local = np.cross(zg_local, xg_local)

        return xg_local, yg_local, zg_local

    def get_3D_coordinates(self,
                           theta: Union[float, np.ndarray] = None
                           ) -> Tuple[Union[float, np.ndarray]]:
        """
        Samples points from the perimeter of this FuselageXSec.

        Args:

            theta: Coordinate in the tangential-ish direction to sample points at. Given in the 2D FuselageXSec
            coordinate system, where:

                * y_2D points along the (global) y_g
                * z_2D points along the (global) z_g

                In other words, a value of:

                    * theta=0     -> samples points from the right side of the FuselageXSec
                    * theta=pi/2  -> samples points from the top of the FuselageXSec
                    * theta=pi    -> samples points from the left side of the FuselageXSec
                    * theta=3pi/2 -> samples points from the bottom of the FuselageXSec

        Returns: Points sampled from the perimeter of the FuselageXSec, as a [x, y, z] tuple.

            If theta is a float, then each of x, y, and z will be floats.

            If theta is an array, then x, y, and z will also be arrays of the same size.

        """
        ### Set defaults
        if theta is None:
            theta = np.linspace(
                0,
                2 * np.pi,
                60 + 1
            )[:-1]

        st = np.sin(np.mod(theta, 2 * np.pi))
        ct = np.cos(np.mod(theta, 2 * np.pi))

        y = (self.width / 2) * np.abs(ct) ** (2 / self.shape) * np.where(ct > 0, 1, -1)
        z = (self.height / 2) * np.abs(st) ** (2 / self.shape) * np.where(st > 0, 1, -1)

        xg_local, yg_local, zg_local = self.compute_frame()

        return (
            self.xyz_c[0] + y * yg_local[0] + z * zg_local[0],
            self.xyz_c[1] + y * yg_local[1] + z * zg_local[1],
            self.xyz_c[2] + y * yg_local[2] + z * zg_local[2],
        )

    def equivalent_radius(self,
                          preserve="area"
                          ) -> float:
        """
        Computes an equivalent radius for non-circular cross-sections. This may be necessary when doing analysis that
        uses axisymmetric assumptions.

        Can either hold area or perimeter fixed, depending on whether cross-sectional area or wetted area is more
        important.

        Args:

            preserve: One of:

                * "area": holds the cross-sectional area constant

                * "perimeter": holds the cross-sectional perimeter (i.e., the wetted area of the Fuselage) constant

        Returns: An equivalent radius value.

        """
        if preserve == "area":
            return (self.xsec_area() / np.pi + 1e-16) ** 0.5
        elif preserve == "perimeter":
            return (self.xsec_perimeter() / (2 * np.pi))
        else:
            raise ValueError("Bad value of `preserve`!")

    def translate(self,
                  xyz: Union[np.ndarray, List[float]]
                  ) -> "FuselageXSec":
        """
        Returns a copy of this FuselageXSec that has been translated by `xyz`.

        Args:
            xyz: The amount to translate the FuselageXSec. Given as a 3-element NumPy vector.

        Returns: A copy of this FuselageXSec, translated by `xyz`.

        """
        new_xsec = copy.copy(self)
        new_xsec.xyz_c = new_xsec.xyz_c + np.array(xyz)
        return new_xsec


if __name__ == '__main__':
    fuse = Fuselage(
        xsecs=[
            FuselageXSec(
                xyz_c=[0, 0, 1],
                radius=0,
            ),
            FuselageXSec(
                xyz_c=[1, 0, 1],
                width=0.5,
                height=0.2,
                shape=5
            ),
            FuselageXSec(
                xyz_c=[2, 0, 1],
                radius=0.2,
            )
        ]
    ).translate([0, 0, 2])
    fuse.draw()



from aerosandbox.dynamics.point_mass.common_point_mass import _DynamicsPointMassBaseClass
from aerosandbox.weights.mass_properties import MassProperties
import aerosandbox.numpy as np
from typing import Union, Dict, Tuple


class DynamicsPointMass3DCartesian(_DynamicsPointMassBaseClass):
    """
    Dynamics instance:
    * simulating a point mass
    * in 3D
    * with velocity parameterized in Cartesian coordinates

    State variables:
        x_e: x-position, in Earth axes. [meters]
        y_e: y-position, in Earth axes. [meters]
        z_e: z-position, in Earth axes. [meters]
        u_e: x-velocity, in Earth axes. [m/s]
        v_e: v-velocity, in Earth axes. [m/s]
        w_e: z-velocity, in Earth axes. [m/s]

    Indirect control variables:
        alpha: Angle of attack. [degrees]
        beta: Sideslip angle. [degrees]
        bank: Bank angle. [radians]

    Control variables:
        Fx_e: Force along the Earth-x axis. [N]
        Fy_e: Force along the Earth-y axis. [N]
        Fz_e: Force along the Earth-z axis. [N]

    """

    def __init__(self,
                 mass_props: MassProperties = None,
                 x_e: Union[float, np.ndarray] = 0,
                 y_e: Union[float, np.ndarray] = 0,
                 z_e: Union[float, np.ndarray] = 0,
                 u_e: Union[float, np.ndarray] = 0,
                 v_e: Union[float, np.ndarray] = 0,
                 w_e: Union[float, np.ndarray] = 0,
                 alpha: Union[float, np.ndarray] = 0,
                 beta: Union[float, np.ndarray] = 0,
                 bank: Union[float, np.ndarray] = 0,
                 ):
        # Initialize state variables
        self.mass_props = MassProperties() if mass_props is None else mass_props
        self.x_e = x_e
        self.y_e = y_e
        self.z_e = z_e
        self.u_e = u_e
        self.v_e = v_e
        self.w_e = w_e

        # Initialize indirect control variables
        self.alpha = alpha
        self.beta = beta
        self.bank = bank

        # Initialize control variables
        self.Fx_e = 0
        self.Fy_e = 0
        self.Fz_e = 0

    @property
    def state(self) -> Dict[str, Union[float, np.ndarray]]:
        return {
            "x_e": self.x_e,
            "y_e": self.y_e,
            "z_e": self.z_e,
            "u_e": self.u_e,
            "v_e": self.v_e,
            "w_e": self.w_e,
        }

    @property
    def control_variables(self) -> Dict[str, Union[float, np.ndarray]]:
        return {
            "alpha": self.alpha,
            "beta" : self.beta,
            "bank" : self.bank,
            "Fx_e" : self.Fx_e,
            "Fy_e" : self.Fy_e,
            "Fz_e" : self.Fz_e,
        }

    def state_derivatives(self) -> Dict[str, Union[float, np.ndarray]]:
        return {
            "x_e": self.u_e,
            "y_e": self.v_e,
            "z_e": self.w_e,
            "u_e": self.Fx_e / self.mass_props.mass,
            "v_e": self.Fy_e / self.mass_props.mass,
            "w_e": self.Fz_e / self.mass_props.mass,
        }

    @property
    def speed(self) -> float:
        return (
                self.u_e ** 2 +
                self.v_e ** 2 +
                self.w_e ** 2
        ) ** 0.5

    @property
    def gamma(self):
        """
        Returns the flight path angle, in radians.

        Positive flight path angle indicates positive vertical speed.

        """
        return np.arctan2(
            -self.w_e,
            (
                    self.u_e ** 2 +
                    self.v_e ** 2
            ) ** 0.5
        )

    @property
    def track(self):
        """
        Returns the track angle, in radians.

        * Track of 0 == North == aligned with x_e axis
        * Track of np.pi / 2 == East == aligned with y_e axis

        """
        return np.arctan2(
            self.v_e,
            self.u_e,
        )

    def convert_axes(self,
                     x_from: float,
                     y_from: float,
                     z_from: float,
                     from_axes: str,
                     to_axes: str,
                     ) -> Tuple[float, float, float]:
        if from_axes == to_axes:
            return x_from, y_from, z_from

        if not (from_axes == "earth" and to_axes == "earth"):
            rot_w_to_e = np.rotation_matrix_from_euler_angles(
                roll_angle=self.bank,
                pitch_angle=self.gamma,
                yaw_angle=self.track,
                as_array=False
            )

        if from_axes == "earth":
            x_e = x_from
            y_e = y_from
            z_e = z_from
        elif from_axes == "wind":
            x_e = rot_w_to_e[0][0] * x_from + rot_w_to_e[0][1] * y_from + rot_w_to_e[0][2] * z_from
            y_e = rot_w_to_e[1][0] * x_from + rot_w_to_e[1][1] * y_from + rot_w_to_e[1][2] * z_from
            z_e = rot_w_to_e[2][0] * x_from + rot_w_to_e[2][1] * y_from + rot_w_to_e[2][2] * z_from
        else:
            x_w, y_w, z_w = self.op_point.convert_axes(
                x_from, y_from, z_from,
                from_axes=from_axes, to_axes="wind"
            )
            x_e = rot_w_to_e[0][0] * x_w + rot_w_to_e[0][1] * y_w + rot_w_to_e[0][2] * z_w
            y_e = rot_w_to_e[1][0] * x_w + rot_w_to_e[1][1] * y_w + rot_w_to_e[1][2] * z_w
            z_e = rot_w_to_e[2][0] * x_w + rot_w_to_e[2][1] * y_w + rot_w_to_e[2][2] * z_w

        if to_axes == "earth":
            x_to = x_e
            y_to = y_e
            z_to = z_e
        elif to_axes == "wind":
            x_to = rot_w_to_e[0][0] * x_e + rot_w_to_e[1][0] * y_e + rot_w_to_e[2][0] * z_e
            y_to = rot_w_to_e[0][1] * x_e + rot_w_to_e[1][1] * y_e + rot_w_to_e[2][1] * z_e
            z_to = rot_w_to_e[0][2] * x_e + rot_w_to_e[1][2] * y_e + rot_w_to_e[2][2] * z_e
        else:
            x_w = rot_w_to_e[0][0] * x_e + rot_w_to_e[1][0] * y_e + rot_w_to_e[2][0] * z_e
            y_w = rot_w_to_e[0][1] * x_e + rot_w_to_e[1][1] * y_e + rot_w_to_e[2][1] * z_e
            z_w = rot_w_to_e[0][2] * x_e + rot_w_to_e[1][2] * y_e + rot_w_to_e[2][2] * z_e
            x_to, y_to, z_to = self.op_point.convert_axes(
                x_w, y_w, z_w,
                from_axes="wind", to_axes=to_axes
            )

        return x_to, y_to, z_to

    def add_force(self,
                  Fx: Union[float, np.ndarray] = 0,
                  Fy: Union[float, np.ndarray] = 0,
                  Fz: Union[float, np.ndarray] = 0,
                  axes="earth",
                  ) -> None:
        Fx_e, Fy_e, Fz_e = self.convert_axes(
            x_from=Fx,
            y_from=Fy,
            z_from=Fz,
            from_axes=axes,
            to_axes="earth"
        )
        self.Fx_e = self.Fx_e + Fx_e
        self.Fy_e = self.Fy_e + Fy_e
        self.Fz_e = self.Fz_e + Fz_e


if __name__ == '__main__':
    dyn = DynamicsPointMass3DCartesian()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_fuselage_aerodynamics_optimization():
    opti = asb.Opti()

    alpha = opti.variable(init_guess=0, lower_bound=0, upper_bound=30)
    beta = opti.variable(init_guess=0)

    fuselage = asb.Fuselage(
        xsecs=[
            asb.FuselageXSec(
                xyz_c=[xi, 0, 0],
                radius=asb.Airfoil("naca0010").local_thickness(0.8 * xi)
            )
            for xi in np.cosspace(0, 1, 20)
        ],
    )

    aero = asb.AeroBuildup(
        airplane=asb.Airplane(
            fuselages=[fuselage]
        ),
        op_point=asb.OperatingPoint(
            velocity=10,
            alpha=alpha,
            beta=beta
        )
    ).run()

    opti.minimize(-aero["L"] / aero["D"])
    sol = opti.solve(verbose=True)
    print(sol.value(alpha))
    assert sol.value(alpha) > 10 and sol.value(alpha) < 20
    assert sol.value(beta) == pytest.approx(0, abs=1e-3)

    opti.minimize(aero["D"])
    sol = opti.solve(verbose=False)
    assert sol.value(alpha) == pytest.approx(0, abs=1e-2)
    assert sol.value(beta) == pytest.approx(0, abs=1e-2)


if __name__ == '__main__':
    test_fuselage_aerodynamics_optimization()
    pytest.main()



import matplotlib.pyplot as plt
import seaborn as sns
import aerosandbox.numpy as np
import aerosandbox.library.aerodynamics as aero


def plot_Cf_flat_plates():
    from aerosandbox.tools.pretty_plots import plt, show_plot

    Res = np.geomspace(1e3, 1e8, 500)
    for method in [
        "blasius",
        "turbulent",
        "hybrid-cengel",
        "hybrid-schlichting",
        "hybrid-sharpe-convex",
        "hybrid-sharpe-nonconvex",
    ]:
        plt.loglog(
            Res,
            aero.Cf_flat_plate(Res, method=method),
            label=method
        )
    plt.ylim(1e-3, 1e-1)
    show_plot(
        "Models for Mean Skin Friction Coefficient of Flat Plate",
        r"$Re$",
        r"$C_f$",
    )


if __name__ == '__main__':
    plot_Cf_flat_plates()



import numpy as _onp
import casadi as _cas
from typing import Any


def is_casadi_type(
        object: Any,
        recursive: bool = True
) -> bool:
    """
    Returns a boolean of whether an object is a CasADi data type or not. If the recursive flag is True,
    iterates recursively, returning True if any subelement (at any depth) is a CasADi type.

    Args:

        object: The object to evaluate.

        recursive: If the object is a list or tuple, recursively iterate through every subelement. If any of the
        subelements (at any depth) are a CasADi type, return True. Otherwise, returns False.

    Returns: A boolean if the object is (or contains, if recursive=True) a CasADi data type.

    """
    t = type(object)

    # NumPy arrays cannot be or contain CasADi types, unless they are object arrays
    if t == _onp.ndarray and object.dtype != 'O':
        return False

    # Skip certain Python types known not to be or contain CasADi types.
    for type_to_skip in (
            float, int, complex,
            bool, str,
            range,
            type(None),
            bytes, bytearray, memoryview
    ):
        if t == type_to_skip:
            return False

    # If it's directly a CasADi type, we're done.
    if (
            t == _cas.MX or
            t == _cas.DM or
            t == _cas.SX
    ):
        return True

    # At this point, we know it's not a CasADi type, but we don't know if it *contains* a CasADi type (relevant if recursing)
    if recursive:
        if (
                issubclass(t, list) or
                issubclass(t, tuple) or
                issubclass(t, set) or
                (
                        t == _onp.ndarray and object.dtype == 'O'
                )
        ):
            for element in object:
                if is_casadi_type(element, recursive=True):
                    return True
            return False

        if issubclass(t, dict):
            for kv in object.items():
                if is_casadi_type(kv, recursive=True):
                    return True
            return False

        return False
    else:
        return False


def is_iterable(x):
    """
    Returns a boolean of whether an object is iterable or not.
    Args:
        x:

    Returns:

    """
    try:
        iter(x)
        return True
    except TypeError:
        return False



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest
import os

"""
Minimization over a simple unimodal function is used here:
    Minimize (x-1) ** 2 + (y-1) ** 2
"""


def sumsqr(x):
    return np.sum(x ** 2)


def test_opti():
    opti = asb.Opti()  # set up an optimization environment

    # Define optimization variables
    x = opti.variable(init_guess=0)
    y = opti.variable(init_guess=0)

    # Define objective
    f = (x - 1) ** 2 + (y - 1) ** 2
    opti.minimize(f)

    # Optimize
    sol = opti.solve()

    for i in [x, y]:
        assert sol.value(i) == pytest.approx(1, abs=1e-4)


def test_save_opti(tmp_path):
    temp_filename = tmp_path / "temp.json"

    opti = asb.Opti(cache_filename=temp_filename)  # set up an optimization environment

    # Define optimization variables
    x = opti.variable(init_guess=0)
    y = opti.variable(init_guess=0)

    # Define objective
    f = (x - 1) ** 2 + (y - 1) ** 2
    opti.minimize(f)

    # Optimize
    sol = opti.solve()

    opti.save_solution()


def test_save_and_load_opti(tmp_path):
    temp_filename = tmp_path / "temp.json"

    ### Round 1 optimization: free optimization

    opti = asb.Opti(
        cache_filename=temp_filename,
        save_to_cache_on_solve=True,
    )
    x = opti.variable(init_guess=0, category="Cat 1")
    y = opti.variable(init_guess=0, category="Cat 2")
    f = (x - 1) ** 2 + (y - 1) ** 2
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    opti.save_solution()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(1)
    assert sol.value(y) == pytest.approx(1)
    assert sol.value(f) == pytest.approx(0)

    ### Round 2 optimization: Cat 1 is fixed from before; slightly different objective now
    opti = asb.Opti(
        cache_filename=temp_filename,
        variable_categories_to_freeze=["Cat 1"],
        load_frozen_variables_from_cache=True,

    )
    x = opti.variable(init_guess=0, category="Cat 1")
    y = opti.variable(init_guess=0, category="Cat 2")
    f = (x - 2) ** 2 + (y - 2) ** 2
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(1)
    assert sol.value(y) == pytest.approx(2)
    assert sol.value(f) == pytest.approx(1)


def test_save_and_load_opti_uncategorized(tmp_path):
    temp_filename = tmp_path / "temp.json"

    ### Round 1 optimization: free optimization

    opti = asb.Opti(
        cache_filename=temp_filename,
        save_to_cache_on_solve=True,
    )
    x = opti.variable(init_guess=0)
    y = opti.variable(init_guess=0)
    f = (x - 1) ** 2 + (y - 1) ** 2
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    opti.save_solution()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(1)
    assert sol.value(y) == pytest.approx(1)
    assert sol.value(f) == pytest.approx(0)

    ### Round 2 optimization: Cat 1 is fixed from before; slightly different objective now
    opti = asb.Opti(
        cache_filename=temp_filename,
        variable_categories_to_freeze=["Uncategorized"],
        load_frozen_variables_from_cache=True,

    )
    x = opti.variable(init_guess=0)
    y = opti.variable(init_guess=0)
    f = (x - 2) ** 2 + (y - 2) ** 2
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(1)
    assert sol.value(y) == pytest.approx(1)
    assert sol.value(f) == pytest.approx(2)


def test_save_and_load_opti_vectorized(tmp_path):
    temp_filename = tmp_path / "temp.json"

    ### Round 1 optimization: free optimization

    opti = asb.Opti(
        cache_filename=temp_filename,
        variable_categories_to_freeze=[],
        save_to_cache_on_solve=True,
    )
    x = opti.variable(init_guess=0, n_vars=3, category="Cat 1")
    y = opti.variable(init_guess=0, n_vars=3, category="Cat 2")
    f = sumsqr(x - 1) + sumsqr(y - 2)
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    opti.save_solution()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(1)
    assert sol.value(y) == pytest.approx(2)
    assert sol.value(f) == pytest.approx(0)

    ### Round 2 optimization: Cat 1 is fixed from before; slightly different objective now
    opti = asb.Opti(
        cache_filename=temp_filename,
        variable_categories_to_freeze=["Cat 1"],
        load_frozen_variables_from_cache=True,

    )
    x = opti.variable(init_guess=0, n_vars=3, category="Cat 1")
    y = opti.variable(init_guess=0, n_vars=3, category="Cat 2")
    f = sumsqr(x - 3) + sumsqr(y - 4)
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(1)
    assert sol.value(y) == pytest.approx(4)
    assert sol.value(f) == pytest.approx(12)


def test_save_and_load_opti_freeze_override(tmp_path):
    temp_filename = tmp_path / "temp.json"

    ### Round 1 optimization: free optimization

    opti = asb.Opti(
        cache_filename=temp_filename,
        variable_categories_to_freeze=[],
        save_to_cache_on_solve=True,
    )
    x = opti.variable(init_guess=0, category="Cat 1")
    y = opti.variable(init_guess=0, category="Cat 2")
    f = (x - 1) ** 2 + (y - 1) ** 2
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    opti.save_solution()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(1)
    assert sol.value(y) == pytest.approx(1)
    assert sol.value(f) == pytest.approx(0)

    ### Round 2 optimization: Cat 1 is fixed from before but then overridden; slightly different objective now
    opti = asb.Opti(
        cache_filename=temp_filename,
        variable_categories_to_freeze=["Cat 1"],
        load_frozen_variables_from_cache=True,

    )
    x = opti.variable(init_guess=3, category="Cat 1", freeze=True)
    y = opti.variable(init_guess=0, category="Cat 2")
    f = (x - 2) ** 2 + (y - 2) ** 2
    opti.minimize(f)

    # Optimize, save to cache, print
    sol = opti.solve()
    for i in ["x", "y", "f"]:
        print(f"{i}: {sol.value(eval(i))}")

    # Test
    assert sol.value(x) == pytest.approx(3)
    assert sol.value(y) == pytest.approx(2)
    assert sol.value(f) == pytest.approx(1)


if __name__ == '__main__':
    # from pathlib import Path
    # tmp_path = Path.home() / "Downloads" / "test"
    # test_save_and_load_opti(tmp_path)
    pytest.main()



import importlib
import sys


def lazy_import(name):
    """
    Allows you to lazily import a module.

    Usage:
    >>> asb = lazy_import("aerosandbox")  # Runs instantly
    >>> asb.Airplane()  # When this is called, ASB will be loaded.

    Args:
        name: The package name

    Returns: The module itself, to be loaded on first use.

    """
    spec = importlib.util.find_spec(name)
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module



import matplotlib.pyplot as plt
import matplotlib as mpl
import aerosandbox.numpy as np
from typing import Tuple, Dict, Union, Callable, List
from scipy import interpolate


def plot_smooth(
        *args,
        color=None,
        label=None,
        resample_resolution: int = 500,
        **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Plots a curve that interpolates a 2D dataset. Same as matplotlib.pyplot.plot(), with the following changes:
        * uses B-splines to draw a smooth curve rather than a jagged polyline
        * By default, plots in line format `fmt='.-'` rather than `fmt='-'`.

    Other than that, almost all matplotlib.pyplot.plot() syntax can be used. See syntax here:
    https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.plot.html

    Example usage:
        >>> import aerosandbox.numpy as np
        >>>
        >>> t = np.linspace(0, 1, 12)  # Parametric variable
        >>> x = np.cos(2 * np.pi * t)
        >>> y = np.cos(2 * np.pi * t ** 4) - t
        >>>
        >>> plot_smooth(
        >>>     x, y, 'o--', color='purple'
        >>> )
        >>> plt.show()

    * Note: a true 2D interpolation is performed - it is not assumed y is a function of x, or vice versa. This can,
    in rare cases, cause single-valuedness to not be preserved in cases where it logically should. If this is the
    case, you need to perform the interpolation yourself without `plot_smooth()`.

    Args:

        *args: Same arguments as `matplotlib.pyplot.plot()`.
            Notes on standard plot() syntax:

                Call signatures:
                >>> plot([x], y, [fmt], *, data=None, **kwargs)
                >>> plot([x], y, [fmt], [x2], y2, [fmt2], ..., **kwargs)

                Examples:
                >>> plot(x, y)        # plot x and y using default line style and color
                >>> plot(x, y, 'bo')  # plot x and y using blue circle markers
                >>> plot(y)           # plot y using x as index array 0..N-1
                >>> plot(y, 'r+')     # ditto, but with red plusses

        color: Specifies the color of any line and/or markers that are plotted (as determined by the `fmt`).

        label: Attaches a label to this line. Use `plt.legend()` to display.

        resample_resolution: The number of points to use when resampling the interpolated curve.

        **kwargs: Same keyword arguments as `matplotlib.pyplot.plot()`.

    Returns: A tuple `(x, y)` of the resampled points on the interpolated curve. Both `x` and `y` are 1D ndarrays.

    """
    argslist = list(args)

    if len(args) == 3:
        x = argslist.pop(0)
        y = argslist.pop(0)
        fmt = argslist.pop(0)
    elif len(args) == 2:
        if isinstance(args[1], str):
            x = np.arange(np.length(args[0]))
            y = argslist.pop(0)
            fmt = argslist.pop(0)
        else:
            x = argslist.pop(0)
            y = argslist.pop(0)
            fmt = '.-'
    elif len(args) == 1:
        x = np.arange(np.length(args[0]))
        y = argslist.pop(0)
        fmt = '.-'
    elif len(args) == 0:
        raise ValueError("Missing plot data. Use syntax `plot_smooth(x, y, fmt, *args, **kwargs)'.")

    bspline = interpolate.make_interp_spline(
        x=np.linspace(0, 1, np.length(y)),
        y=np.stack(
            (x, y), axis=1
        )
    )
    result = bspline(np.linspace(0, 1, resample_resolution))
    x_resample = result[:, 0]
    y_resample = result[:, 1]

    scatter_kwargs = {
        **kwargs,
        'linewidth': 0,
    }
    if color is not None:
        scatter_kwargs['color'] = color

    line, = plt.plot(
        x,
        y,
        fmt,
        *argslist,
        **scatter_kwargs
    )

    if color is None:
        color = line.get_color()

    line_kwargs = {
        'color'     : color,
        'label'     : label,
        **kwargs,
        'markersize': 0,
    }

    plt.plot(
        x_resample,
        y_resample,
        fmt,
        *argslist,
        **line_kwargs
    )

    return x_resample, y_resample



from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.conventional import airplane
import aerosandbox as asb
import aerosandbox.numpy as np

for i, wing in enumerate(airplane.wings):
    for j, xsec in enumerate(wing.xsecs):
        af = xsec.airfoil
        af.generate_polars(
            cache_filename=f"cache/{af.name}.json",
            xfoil_kwargs=dict(
                xfoil_repanel="naca" in af.name.lower()
            )
        )
        airplane.wings[i].xsecs[j].airfoil = af

op_point = asb.OperatingPoint(
    velocity=25,
    alpha=3
)

vlm = asb.VortexLatticeMethod(
    airplane,
    op_point,
    align_trailing_vortices_with_wind=True,
    chordwise_resolution=12,
    spanwise_resolution=12,
)
vlm_aero = vlm.run()

ab = asb.AeroBuildup(
    airplane,
    op_point
)
ab_aero = ab.run()

avl = asb.AVL(
    airplane,
    op_point,
)
avl_aero = avl.run()

for k, v in {
    "VLM": vlm_aero,
    "AVL": avl_aero,
    "AB" : ab_aero
}.items():
    print(f"{k}:")
    for f in ["CL", "CD", "Cm"]:
        print(f"\t{f} : {v[f]}")
    print(f"\tL/D : {v['CL'] / v['CD']}")



from aerosandbox.tools.pretty_plots.utilities.natural_univariate_spline import NaturalUnivariateSpline as Spline
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm


def plot_with_bootstrapped_uncertainty(
        x: np.ndarray,
        y: np.ndarray,
        y_stdev: float,
        ci: float = 0.95,
        color=None,
        draw_line=True,
        draw_ci=True,
        draw_data=True,
        label_line=None,
        label_ci=None,
        label_data=None,
        n_bootstraps=2000,
        n_fit_points=300,
        spline_degree=3,
        normalize=True,
):

    if not (ci > 0 and ci < 1):
        raise ValueError("Confidence interval `ci` should be in the range of (0, 1).")

    ### Discard any NaN points
    isnan = np.logical_or(
        np.isnan(x),
        np.isnan(y),
    )
    x = x[~isnan]
    y = y[~isnan]

    ### Prepare for normalization
    if normalize:
        x_min = np.min(x)
        x_rng = np.max(x) - x_min
        y_min = np.min(y)
        y_rng = np.max(y) - y_min

        x_normalize = lambda xi: (xi - x_min) / x_rng
        y_normalize = lambda yi: (yi - y_min) / y_rng
        x_unnormalize = lambda xi: xi * x_rng + x_min
        y_unnormalize = lambda yi: yi * y_rng + y_min

    else:
        x_normalize = lambda xi: xi
        y_normalize = lambda yi: yi
        x_unnormalize = lambda xi: xi
        y_unnormalize = lambda yi: yi

    ### Prepare for the bootstrap
    x_fit = np.linspace(
        np.min(x),
        np.max(x),
        n_fit_points
    )

    y_bootstrap_fits = np.empty((n_bootstraps, len(x_fit)))

    for i in tqdm(range(n_bootstraps), desc="Bootstrapping", unit=" samples"):

        ### Obtain a bootstrap resample
        ### Here, instead of truly resampling, we just pick weights that effectively mimic a resample.
        ### A computationally-efficient way to pick weights is the following clever trick with uniform sampling:
        splits = np.random.rand(len(x) + 1) * len(x)  # "limit" bootstrapping
        splits[0] = 0
        splits[-1] = len(x)

        weights = np.diff(np.sort(splits))

        y_bootstrap_fits[i, :] = y_unnormalize(Spline(
            x=x_normalize(x),
            y=y_normalize(y),
            w=weights / y_normalize(y_stdev),
            s=len(x) * (y_normalize(y).max() - y_normalize(y).min()) ** 0.5,
            k=spline_degree,
            ext='extrapolate'
        )(x_normalize(x_fit)))

    ### Compute a confidence interval using equal-tails method
    y_median_and_ci = np.nanquantile(
        y_bootstrap_fits,
        q=[
            (1 - ci) / 2,
            0.5,
            1 - (1 - ci) / 2
        ],
        axis=0
    )

    if draw_line:
        line, = plt.plot(
            x_fit,
            y_median_and_ci[1, :],
            color=color,
            label=label_line
        )
        if color is None:
            color = line.get_color()

    if draw_ci:
        plt.fill_between(
            x_fit,
            y_median_and_ci[0, :],
            y_median_and_ci[2, :],
            color=color,
            label=label_ci,
            alpha=0.25,
            linewidth=0
        )
    if draw_data:
        line, = plt.plot(
            x,
            y,
            ".",
            color=color,
            label=label_data,
            alpha=0.5
        )
        if color is None:
            color = line.get_color()
    return x_fit, y_bootstrap_fits


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    np.random.seed(0)

    ### Generate data
    x = np.linspace(0, 10, 101)
    y_true = np.abs(x - 5)  # np.sin(x)
    y_noisy = y_true + 0.1 * np.random.randn(len(x))

    ### Plot spline regression
    fig, ax = plt.subplots(dpi=300)
    x_fit, y_bootstrap_fits = plot_with_bootstrapped_uncertainty(
        x,
        y_noisy,
        y_stdev=0.1 * np.std(y_noisy),
        label_line="Best Estimate",
        label_data="Data",
        label_ci="95% CI",
    )
    ax.plot(x, y_true, "k", label="True Function", alpha=0.2)

    p.show_plot(
        "Spline Bootstrapping Test",
        r"$x$",
        r"$y$",
    )



import aerosandbox.numpy as np
from typing import Union


def CDA_control_linkage(
        Re_l: Union[float, np.ndarray],
        linkage_length: Union[float, np.ndarray],
        is_covered: Union[bool, np.ndarray] = False,
        is_top: Union[bool, np.ndarray] = False,
) -> Union[float, np.ndarray]:
    """
    Computes the drag coefficient of a typical control usage as used on a well-manufactured RC airplane.

    See study with original data at `AeroSandbox/studies/LinkageDrag`.

    Data from:
        * Hepperle, Martin. "Drag of Linkages". https://www.mh-aerotools.de/airfoils/linkage.htm
        * Summarizes data from "Werner Würz, published in the papers of the ISF-Seminar in December 1989 in Baden, Switzerland."

    Args:

        Re_l: Reynolds number, with reference length as the length of the linkage.

        linkage_length: The length of the linkage. [m]

        is_covered: A boolean of whether an aerodynamic fairing is placed around the linkage.

        is_top: A boolean of whether the linkage is on the top surface of the wing (True) or the bottom surface (
        False). Differences in local boundary layer and inviscid effects cause local velocity changes.

    Returns: The drag area [m^2] of the control linkage.

    """
    x = dict(
        Re_l=Re_l,
        linkage_length=linkage_length,
        is_covered=is_covered,
        is_top=is_top
    )

    p = {
        'CD0'               : 7.833083680086374e-05,
        'CD1'               : 0.0001216877860785463,
        'c_length'          : 30.572471745477774,
        'covered_drag_ratio': 0.7520722978405192,
        'top_drag_ratio'    : 1.1139040832208857
    }

    Re = x["Re_l"]
    linkage_length = x["linkage_length"]
    is_covered = x["is_covered"]
    is_top = x["is_top"]

    side_drag_multiplier = np.where(
        is_top,
        p["top_drag_ratio"],
        1
    )
    covered_drag_multiplier = np.where(
        is_covered,
        p["covered_drag_ratio"],
        1
    )
    linkage_length_multiplier = 1 + p["c_length"] * linkage_length

    CDA_raw = (
            p["CD1"] / (Re / 1e5) +
            p["CD0"]
    )

    return side_drag_multiplier * covered_drag_multiplier * linkage_length_multiplier * CDA_raw



import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from typing import Tuple, Dict, Union, Callable, List
from scipy import interpolate
from aerosandbox.tools.string_formatting import eng_string


def contour(
        *args,
        levels: Union[int, List, np.ndarray] = 31,
        colorbar: bool = True,
        linelabels: bool = True,
        cmap=None,
        alpha: float = 0.7,
        extend: str = "neither",
        linecolor="k",
        linewidths: float = 0.5,
        extendrect: bool = True,
        linelabels_format: Union[str, Callable[[float], str]] = eng_string,
        linelabels_fontsize: float = 8,
        colorbar_label: str = None,
        z_log_scale: bool = False,
        contour_kwargs: Dict = None,
        contourf_kwargs: Dict = None,
        colorbar_kwargs: Dict = None,
        linelabels_kwargs: Dict = None,
        **kwargs,
):
    """
    An analogue for plt.contour and plt.tricontour and friends that produces a much prettier default graph.

    Can take inputs with either contour or tricontour syntax.

    See syntax here:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.contour.html
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.contourf.html
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontourf.html

    Args:
        X: See contour docs.

        Y: See contour docs.

        Z: See contour docs.

        levels: See contour docs.

        colorbar: Should we draw a colorbar?

        linelabels: Should we add line labels?

        cmap: What colormap should we use?

        alpha: What transparency should all plot elements be?

        extend: See contour docs.

        linecolor: What color should the line labels be?

        linewidths: See contour docs.

        extendrect: See colorbar docs.

        linelabels_format: See ax.clabel docs.

        linelabels_fontsize: See ax.clabel docs.

        contour_kwargs: Additional keyword arguments for contour.

        contourf_kwargs: Additional keyword arguments for contourf.

        colorbar_kwargs: Additional keyword arguments for colorbar.

        linelabels_kwargs: Additional keyword arguments for the line labels (ax.clabel).

        **kwargs: Additional keywords, which are passed to both contour and contourf.


    Returns: A tuple of (contour, contourf, colorbar) objects.

    """
    bad_signature_error = ValueError("Call signature should be one of:\n"
                                     "  * `contour(Z, **kwargs)`\n"
                                     "  * `contour(X, Y, Z, **kwargs)`\n"
                                     "  * `contour(X, Y, Z, levels, **kwargs)`"
                                     )

    ### Parse *args
    if len(args) == 1:
        X = None
        Y = None
        Z = args[0]
    elif len(args) == 3:
        X = args[0]
        Y = args[1]
        Z = args[2]
    else:
        raise bad_signature_error
    if X is None:
        X = np.arange(Z.shape[1])
    if Y is None:
        Y = np.arange(Z.shape[0])

    ### Set defaults
    if cmap is None:
        cmap = mpl.colormaps.get_cmap('viridis')
    if contour_kwargs is None:
        contour_kwargs = {}
    if contourf_kwargs is None:
        contourf_kwargs = {}
    if colorbar_kwargs is None:
        colorbar_kwargs = {}
    if linelabels_kwargs is None:
        linelabels_kwargs = {}

    shared_kwargs = kwargs
    if levels is not None:
        shared_kwargs["levels"] = levels
    if alpha is not None:
        shared_kwargs["alpha"] = alpha
    if extend is not None:
        shared_kwargs["extend"] = extend
    if z_log_scale:
        if np.any(Z) <= 0:
            raise ValueError(
                "All values of the `Z` input to `contour()` should be nonnegative if `z_log_scale` is True!"
            )

        Z_ratio = np.nanmax(Z) / np.nanmin(Z)
        log10_ceil_z_max = np.ceil(np.log10(np.nanmax(Z)))
        log10_floor_z_min = np.floor(np.log10(np.nanmin(Z)))

        try:
            default_levels = int(levels)
        except TypeError:
            default_levels = 31
        divisions_per_decade = np.ceil(default_levels / np.log10(Z_ratio)).astype(int)

        if Z_ratio > 1e8:
            locator = mpl.ticker.LogLocator()
        else:
            locator = mpl.ticker.LogLocator(
                subs=np.geomspace(1, 10, divisions_per_decade + 1)[:-1]
            )

        shared_kwargs = {
            "norm"   : mpl.colors.LogNorm(),
            "locator": locator,
            **shared_kwargs
        }

        colorbar_kwargs = {
            "norm": mpl.colors.LogNorm(),
            **colorbar_kwargs
        }

    if colorbar_label is not None:
        colorbar_kwargs["label"] = colorbar_label

    contour_kwargs = {
        "colors"    : linecolor,
        "linewidths": linewidths,
        **shared_kwargs,
        **contour_kwargs
    }
    contourf_kwargs = {
        "cmap": cmap,
        **shared_kwargs,
        **contourf_kwargs
    }

    colorbar_kwargs = {
        "extendrect": extendrect,
        **colorbar_kwargs
    }

    linelabels_kwargs = {
        "inline"  : 1,
        "fontsize": linelabels_fontsize,
        "fmt"     : linelabels_format,
        **linelabels_kwargs
    }

    ### Now, with all the kwargs merged, do the actual plotting:

    try:
        cont = plt.contour(X, Y, Z, **contour_kwargs)
        contf = plt.contourf(X, Y, Z, **contourf_kwargs)
    except TypeError as e:
        try:
            cont = plt.tricontour(X, Y, Z, **contour_kwargs)
            contf = plt.tricontourf(X, Y, Z, **contourf_kwargs)
        except TypeError:
            raise e

    if colorbar:
        from matplotlib import cm

        cbar = plt.colorbar(
            ax=contf.axes,
            mappable=cm.ScalarMappable(
                norm=contf.norm,
                cmap=contf.cmap,
            ),
            **colorbar_kwargs
        )

        if z_log_scale:

            if Z_ratio >= 1e2:
                cbar.ax.yaxis.set_major_locator(mpl.ticker.LogLocator())
                cbar.ax.yaxis.set_minor_locator(mpl.ticker.LogLocator(subs=np.arange(10)))
                cbar.ax.yaxis.set_major_formatter(mpl.ticker.LogFormatterSciNotation())
                cbar.ax.yaxis.set_minor_formatter(mpl.ticker.NullFormatter())
            elif Z_ratio >= 10 ** 1.5:
                cbar.ax.yaxis.set_major_locator(mpl.ticker.LogLocator())
                cbar.ax.yaxis.set_minor_locator(mpl.ticker.LogLocator(subs=np.arange(10)))
                cbar.ax.yaxis.set_major_formatter(mpl.ticker.LogFormatterSciNotation())
                cbar.ax.yaxis.set_minor_formatter(mpl.ticker.LogFormatterSciNotation())
            else:
                cbar.ax.yaxis.set_major_locator(mpl.ticker.LogLocator(subs=np.arange(10)))
                cbar.ax.yaxis.set_minor_locator(mpl.ticker.LogLocator(subs=np.arange(100)))
                cbar.ax.yaxis.set_major_formatter(mpl.ticker.ScalarFormatter())
                cbar.ax.yaxis.set_minor_formatter(mpl.ticker.NullFormatter())

    else:
        cbar = None

    if linelabels:
        cont.axes.clabel(cont, **linelabels_kwargs)

    return cont, contf, cbar


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    x = np.linspace(0, 1, 300)
    y = np.linspace(0, 1, 200)
    X, Y = np.meshgrid(x, y)

    Z_ratio = 1.5

    Z = 10 ** (
            Z_ratio / 2 * np.cos(
        2 * np.pi * (X ** 4 + Y ** 4)
    )
    )

    fig, ax = plt.subplots(figsize=(6, 6))

    cmap = plt.get_cmap("rainbow")

    cont, contf, cbar = contour(
        X, Y, Z,
        z_log_scale=True,
        cmap=cmap,
        levels=20,
        colorbar_label="Colorbar label"
    )
    p.show_plot(
        "Title",
        "X label",
        "Y label"
    )



# From https://www.grc.nasa.gov/WWW/K-12/airplane/normal.html

def mach_number_after_normal_shock(
        mach_upstream,
        gamma=1.4,
):
    """
    Computes the mach number immediately after a normal shock wave.

    Args:
        mach_upstream: The mach number immediately before the normal shock wave.
        gamma: The ratio of specific heats of the fluid. 1.4 for air.

    Returns: The mach number immediately after the normal shock wave.

    """
    gm1 = gamma - 1
    m2 = mach_upstream ** 2

    return (
            (gm1 * m2 + 2) / (2 * gamma * m2 - gm1)
    ) ** 0.5


def density_ratio_across_normal_shock(
        mach_upstream,
        gamma=1.4
):
    """
    Computes the ratio of fluid density across a normal shock.

    Specifically, returns: rho_after_shock / rho_before_shock

    Args:
        mach_upstream: The mach number immediately before the normal shock wave.
        gamma: The ratio of specific heats of the fluid. 1.4 for air.

    Returns: rho_after_shock / rho_before_shock

    """
    return (
            (gamma + 1) * mach_upstream ** 2
    ) / (
            (gamma - 1) * mach_upstream ** 2 + 2
    )


def temperature_ratio_across_normal_shock(
        mach_upstream,
        gamma=1.4
):
    """
    Computes the ratio of fluid temperature across a normal shock.

    Specifically, returns: T_after_shock / T_before_shock

    Args:
        mach_upstream: The mach number immediately before the normal shock wave.
        gamma: The ratio of specific heats of the fluid. 1.4 for air.

    Returns: T_after_shock / T_before_shock

    """
    gm1 = gamma - 1
    m2 = mach_upstream ** 2
    return (
            (2 * gamma * m2 - gm1) * (gm1 * m2 + 2)
    ) / (
            (gamma + 1) ** 2 * m2
    )


def pressure_ratio_across_normal_shock(
        mach_upstream,
        gamma=1.4
):
    """
    Computes the ratio of fluid static pressure across a normal shock.

    Specifically, returns: P_after_shock / P_before_shock

    Args:
        mach_upstream: The mach number immediately before the normal shock wave.
        gamma: The ratio of specific heats of the fluid. 1.4 for air.

    Returns: P_after_shock / P_before_shock

    """
    m2 = mach_upstream ** 2
    return (
            2 * gamma * m2 - (gamma - 1)
    ) / (
        (gamma + 1)
    )


def total_pressure_ratio_across_normal_shock(
        mach_upstream,
        gamma=1.4
):
    """
    Computes the ratio of fluid total pressure across a normal shock.

    Specifically, returns: Pt_after_shock / Pt_before_shock

    Args:
        mach_upstream: The mach number immediately before the normal shock wave.
        gamma: The ratio of specific heats of the fluid. 1.4 for air.

    Returns: Pt_after_shock / Pt_before_shock

    """
    return density_ratio_across_normal_shock(
        mach_upstream=mach_upstream,
        gamma=gamma
    ) ** (gamma / (gamma - 1)) * (
            (gamma + 1) / (2 * gamma * mach_upstream ** 2 - (gamma - 1))
    ) ** (1 / (gamma - 1))


if __name__ == '__main__':
    def q_ratio(mach):
        return (
                density_ratio_across_normal_shock(mach) *
                (
                        mach_number_after_normal_shock(mach) *
                        temperature_ratio_across_normal_shock(mach) ** 0.5
                ) ** 2
        )


    q_ratio(2)



from aerosandbox.modeling.interpolation import InterpolatedModel
import pytest
import aerosandbox as asb
import aerosandbox.numpy as np


def underlying_function(x):  # Softmax of three linear functions
    def f1(x):
        return -3 * x - 4

    def f2(x):
        return -0.25 * x + 1

    def f3(x):
        return 2 * x - 12

    return np.softmax(
        f1(x),
        f2(x),
        f3(x),
        hardness=1
    )


def get_interpolated_model(res=51):
    x_samples = np.linspace(-10, 10, res)
    f_samples = underlying_function(x_samples)

    return InterpolatedModel(
        x_data_coordinates=x_samples,
        y_data_structured=f_samples
    )


def plot_underlying_function():
    import matplotlib.pyplot as plt
    import seaborn as sns
    sns.set(palette=sns.color_palette("husl"))
    x = np.linspace(-10, 10, 500)
    f = underlying_function(x)

    fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    plt.plot(x, f)
    plt.title("Underlying Function")
    plt.show()


def plot_interpolated_model(interpolated_model=get_interpolated_model()):
    interpolated_model.plot()


def test_solve_actual_function():
    opti = asb.Opti()

    x = opti.variable(init_guess=-6)

    opti.minimize(underlying_function(x))

    sol = opti.solve()

    assert sol.value(x) == pytest.approx(4.85358, abs=1e-3)


def test_solve_interpolated_unbounded(
        interpolated_model=get_interpolated_model()
):
    opti = asb.Opti()

    x = opti.variable(init_guess=-5)

    opti.minimize(interpolated_model(x))

    sol = opti.solve()

    assert sol.value(x) == pytest.approx(4.85358, abs=0.1)


# def test_solve_interpolated_infeasible_start_but_bounded(
#         interpolated_model=get_interpolated_model()
# ):
#     opti = asb.Opti()
#
#     x = opti.variable(init_guess=-11, lower_bound=-10, upper_bound=10)
#
#     opti.minimize(interpolated_model(x))
#
#     sol = opti.solve()
#
#     assert sol.value(x) == pytest.approx(4.85358, abs=0.1)


if __name__ == '__main__':
    # plot_underlying_function()
    # plot_interpolated_model(interpolated_model())
    pytest.main()



from aerosandbox.common import ExplicitAnalysis
import aerosandbox.numpy as np
import subprocess
from pathlib import Path
from aerosandbox.geometry import Airfoil
from aerosandbox.aerodynamics.aero_3D.avl import AVL
from typing import Union, List, Dict
import tempfile
import warnings
import os
from textwrap import dedent
import shutil


class MSES(ExplicitAnalysis):
    """

    An interface to MSES, MSET, and MPLOT, a 2D airfoil analysis system developed by Mark Drela at MIT.

    Requires compiled binaries for all the programs to be on your computer;
    MSES is available here: https://web.mit.edu/drela/Public/web/mses/
    Academics can get a copy by emailing the MIT Tech. Licensing Office;
    MIT affiliates can find a copy on Athena.

    It is recommended (but not required) that you add MSES, MSET, and MPLOT to your system PATH environment variable
    such that they can be called with the commands `mses`, `mset`, and `mplot`. If this is not the case, you need to
    specify the path to these executables using the command arguments of the constructor.

    -----
    X11 Notes:

    Note that MSES, MSET, and MPLOT by default open up X11 windows on your computer. If you prefer that this doesn't
    happen (for extra speed), or if you cannot have this happen (e.g., you are computing in an environment without
    proper X11 support, like Windows Subsystem for Linux), you should use XVFB. https://en.wikipedia.org/wiki/Xvfb

    XVFB is a virtual "display" server that can receive X11 output and safely dump it. (If you don't use XVFB and you
    don't have proper X11 support on your computer, this AeroSandbox MSES module will simply error out during the
    MSET call - probably not what you want.)

    To install XVFB on a Linux machine, use:

    ```bash
    sudo apt-get install xvfb
    ```

    Then, when instantiating this MSES instance in AeroSandbox, pass the `use_xvfb` flag to be True. Default behavior
    here is that this class will look for the XVFB executable, `xvfb-run`, on your machine. If it finds it,
    it will run with XVFB enabled. If it does not, it will run without XVFB.

    -----

    Usage example:

    >>> ms = MSES(
    >>>     airfoil=Airfoil("naca2412").repanel(n_points_per_side=100),
    >>>     Re=1e6,
    >>>     mach=0.2,
    >>> )
    >>>
    >>> result_at_single_alpha = ms.alpha(5)
    >>> #result_at_several_CLs = ms.cl([0.5, 0.7, 0.8, 0.9])
    >>> result_at_multiple_alphas = ms.alpha([3, 5, 60]) # Note: if a result does not converge (such as the 60 degree case here), it will not be included in the results.

    """

    def __init__(self,
                 airfoil: Airfoil,
                 n_crit: float = 9.,
                 xtr_upper: float = 1.,
                 xtr_lower: float = 1.,
                 max_iter: int = 100,
                 mset_command: str = "mset",
                 mses_command: str = "mses",
                 mplot_command: str = "mplot",
                 use_xvfb: bool = None,
                 xvfb_command: str = "xvfb-run -a",
                 verbosity: int = 1,
                 timeout_mset: Union[float, int, None] = 10,
                 timeout_mses: Union[float, int, None] = 60,
                 timeout_mplot: Union[float, int, None] = 10,
                 working_directory: str = None,
                 behavior_after_unconverged_run: str = "reinitialize",
                 mset_alpha: float = 0,
                 mset_n: int = 141,
                 mset_e: float = 0.4,
                 mset_io: int = 37,
                 mset_x: float = 0.850,
                 mses_mcrit: float = 0.99,
                 mses_mucon: float = -1.0,
                 ):
        """
        Interface to XFoil. Compatible with both XFoil v6.xx (public) and XFoil v7.xx (private, contact Mark Drela at
        MIT for a copy.)

        Args:
            # TODO docs
            airfoil: The angle of attack [degrees]

            Re: The chord-referenced Reynolds number

            mach: The freestream Mach number

            n_crit: The critical Tollmein-Schlichting wave amplification factor

            xtr_upper: The upper-surface trip location [x/c]

            xtr_lower: The lower-surface trip location [x/c]

            full_potential: If this is set True, it will turn full-potential mode on. Note that full-potential mode
            is only available in XFoil v7.xx or higher. (Unless you have specifically gone through the trouble of
            acquiring a copy of XFoil v7.xx you likely have v6.xx. Version 7.xx is not publicly distributed as of
            2022; contact Mark Drela at MIT for a copy.) Note that if you enable this flag with XFoil v6.xx,
            you'll likely get an error (no output file generated).

            max_iter: How many iterations should we let XFoil do?

            xfoil_command: The command-line argument to call XFoil.

                * If XFoil is on your system PATH, then you can just leave this as "xfoil".

                * If XFoil is not on your system PATH, then you should provide a filepath to the XFoil executable.

                Note that XFoil is not on your PATH by default. To tell if XFoil is not on your system PATH,
                open up a terminal and type "xfoil".

                    * If the XFoil menu appears, it's on your PATH.

                    * If you get something like "'xfoil' is not recognized as an internal or external command..." or
                    "Command 'xfoil' not found, did you mean...", then it is not on your PATH and you'll need to
                    specify the location of your XFoil executable as a string.

                To add XFoil to your path, modify your system's environment variables. (Google how to do this for
                your OS.)

            xfoil_repanel: Controls whether to allow XFoil to repanel your airfoil using its internal methods (PANE
            -> PPAR, both with default settings, 160 nodes)

            verbose: Controls whether or not XFoil output is printed to command line.

            timeout: Controls how long any individual XFoil run (i.e. alpha sweep) is allowed to run before the
            process is killed. Given in units of seconds. To disable timeout, set this to None.

            working_directory: Controls which working directory is used for the XFoil input and output files. By
            default, this is set to a TemporaryDirectory that is deleted after the run. However, you can set it to
            somewhere local for debugging purposes.

        """
        if use_xvfb is None:
            trial_run = subprocess.run(
                xvfb_command,  # Analogous to "xvfb-run", perhaps with additional arguments
                capture_output=True,
                shell=True,
                text=True,
            )
            expected_result = 'xvfb-run: usage error:'
            use_xvfb = expected_result in trial_run.stderr or expected_result in trial_run.stdout

        if not use_xvfb:
            xvfb_command = ""

        self.airfoil = airfoil
        self.n_crit = n_crit
        self.xtr_upper = xtr_upper
        self.xtr_lower = xtr_lower
        self.max_iter = max_iter
        self.mset_command = mset_command
        self.mses_command = mses_command
        self.mplot_command = mplot_command
        self.use_xvfb = use_xvfb
        self.xvfb_command = xvfb_command
        self.verbosity = verbosity
        self.timeout_mses = timeout_mses
        self.timeout_mset = timeout_mset
        self.timeout_mplot = timeout_mplot
        self.working_directory = working_directory
        self.behavior_after_unconverged_run = behavior_after_unconverged_run
        self.mset_alpha = mset_alpha
        self.mset_n = mset_n
        self.mset_e = mset_e
        self.mset_io = mset_io
        self.mset_x = mset_x
        self.mses_mcrit = mses_mcrit
        self.mses_mucon = mses_mucon

    def run(self,
            alpha: Union[float, np.ndarray, List] = 0.,
            Re: Union[float, np.ndarray, List] = 0.,
            mach: Union[float, np.ndarray, List] = 0.01,
            ):
        ### Make all inputs iterables:
        alphas, Res, machs = np.broadcast_arrays(
            np.ravel(alpha),
            np.ravel(Re),
            np.ravel(mach),
        )

        # Set up a temporary directory
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            ### Alternatively, work in another directory:
            if self.working_directory is not None:
                directory = Path(self.working_directory)  # For debugging

            # Handle the airfoil file
            airfoil_file = "airfoil.dat"
            self.airfoil.write_dat(directory / airfoil_file)

            def mset(mset_alpha):
                mset_keystrokes = dedent(f"""\
                15
                case
                7
                n {self.mset_n}
                e {self.mset_e}
                i {self.mset_io}
                o {self.mset_io}
                x {self.mset_x}
                
                1
                {mset_alpha}
                2
                
                3
                4
                0
                """)

                if self.verbosity >= 1:
                    print(f"Generating mesh at alpha = {mset_alpha} with MSES...")

                return subprocess.run(
                    f'{self.xvfb_command} "{self.mset_command}" "{airfoil_file}"',
                    input=mset_keystrokes,
                    cwd=directory,
                    capture_output=True,
                    text=True,
                    shell=True,
                    check=True,
                    timeout=self.timeout_mset
                )

            try:
                mset_run = mset(mset_alpha=alphas[0])
            except subprocess.CalledProcessError as e:
                print(e.stdout)
                print(e.stderr)
                if "BadName (named color or font does not exist)" in e.stderr:
                    raise RuntimeError("MSET via AeroSandbox errored becausee it couldn't launch an X11 window.\n"
                                       "Try either installing a typical X11 client, or install Xvfb, which is\n"
                                       "a virtual X11 server. More details in the AeroSandbox MSES docstring.")

            runs_output = {}

            for i, (alpha, mach, Re) in enumerate(zip(alphas, machs, Res)):

                if self.verbosity >= 1:
                    print(f"Solving alpha = {alpha:.3f}, mach = {mach:.4f}, Re = {Re:.3e} with MSES...")

                with open(directory / "mses.case", "w+") as f:
                    f.write(dedent(f"""\
                    3  4  5  7
                    3  4  5  7
                    {mach}   0.0   {alpha} | MACHin  CLIFin  ALFAin
                    3  2                             | ISMOM  IFFBC  [ DOUXin DOUYin SRCEin ]
                    {Re}  {self.n_crit}          | REYNin ACRIT [ KTRTYP ]
                    {self.xtr_lower}    {self.xtr_upper}                   | XTR1 XTR2
                    {self.mses_mcrit}  {self.mses_mucon}                      | MCRIT  MUCON
                    0    0                           | ISMOVE  ISPRES
                    0    0                           | NMODN   NPOSN
                    
                    """))

                mses_keystrokes = dedent(f"""\
                    {self.max_iter}
                    0
                    """)

                mses_run = subprocess.run(
                    f'{self.xvfb_command} "{self.mses_command}" case',
                    input=mses_keystrokes,
                    cwd=directory,
                    capture_output=True,
                    text=True,
                    shell=True,
                    check=True,
                    timeout=self.timeout_mses
                )
                if self.verbosity >= 2:
                    print(mses_run.stdout)
                    print(mses_run.stderr)

                converged = "Converged on tolerance" in mses_run.stdout
                if not converged:
                    if self.behavior_after_unconverged_run == "reinitialize":
                        if self.verbosity >= 1:
                            print("Run did not converge. Reinitializing mesh and continuing...")
                        try:
                            next_alpha = alphas[i + 1]
                        except IndexError:
                            break
                        mset_run = mset(mset_alpha=next_alpha)
                    elif self.behavior_after_unconverged_run == "terminate":
                        if self.verbosity >= 1:
                            print("Run did not converge. Skipping all subsequent runs...")
                            break

                    continue

                mplot_keystrokes = dedent(f"""\
                        1
                        12
                        0
                        0
                    """)

                mplot_run = subprocess.run(
                    f'{self.xvfb_command} "{self.mplot_command}" case',
                    input=mplot_keystrokes,
                    cwd=directory,
                    capture_output=True,
                    text=True,
                    shell=True,
                    check=True,
                    timeout=self.timeout_mplot
                )
                if self.verbosity >= 2:
                    print(mplot_run.stdout)
                    print(mplot_run.stderr)

                raw_output = mplot_run.stdout. \
                    replace("top Xtr", "xtr_top"). \
                    replace("bot Xtr", "xtr_bot"). \
                    replace("at x,y", "x_ac")

                run_output = AVL.parse_unformatted_data_output(raw_output)

                # Merge runs_output and run_output
                for k in run_output.keys():
                    try:
                        runs_output[k].append(
                            run_output[k]
                        )
                    except KeyError:  # List not created yet
                        runs_output[k] = [run_output[k]]

            # Clean up the dictionary
            runs_output = {k: np.array(v) for k, v in runs_output.items()}
            # runs_output["mach"] = runs_output.pop("Ma")
            runs_output = {
                "mach": runs_output.pop("Ma"),
                **runs_output
            }

            return runs_output


if __name__ == '__main__':
    from pathlib import Path
    from pprint import pprint

    ms = MSES(
        airfoil=Airfoil("rae2822"),  # .repanel(n_points_per_side=30),
        working_directory="/mnt/c/Users/peter/Downloads/msestest/",
        # max_iter=120,
        verbosity=1,
        behavior_after_unconverged_run="terminate",
        mset_n=300,
        max_iter=100,
        # verbose=False
    )
    res = ms.run(
        alpha=3,
        mach=np.arange(0.55, 0.8, 0.005),
        # Re=1e6,
    )
    pprint(res)

    import matplotlib

    matplotlib.use("WebAgg")

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots()
    plt.plot(res['mach'], res['CD'], ".-")
    p.show_plot()



import aerosandbox.numpy as np


def mass_turbofan(
        m_dot_core_corrected,
        overall_pressure_ratio,
        bypass_ratio,
        diameter_fan,

):
    """
    Computes the combined mass of a bare turbofan, nacelle, and accessory and pylon weights.

    Bare weight depends on m_dot, OPR, and BPR.

    Nacelle weight is a function of various areas and fan diameter.

    From TASOPT documentation by Mark Drela, available here: http://web.mit.edu/drela/Public/web/tasopt/TASOPT_doc.pdf
        Section: "Turbofan Weight Model from Historical Data"

    Args:

        m_dot_core_corrected: The mass flow of the core only, corrected to standard conditions. [kg/s]

        overall_pressure_ratio: The overall pressure ratio (OPR) [-]

        bypass_ratio: The bypass ratio (BPR) [-]

        diameter_fan: The diameter of the fan. [m]

    Returns: The total engine mass. [kg]

    """
    kg_to_lbm = 2.20462262
    m_to_ft = 1 / 0.3048

    ##### Compute bare turbofan weight
    m_dot_core_corrected_lbm_per_sec = m_dot_core_corrected * kg_to_lbm  # Converts from kg/s to lbm/s

    ### Parameters determined via least-squares fitting by Drela in TASOPT doc.
    b_m = 1
    b_pi = 1
    b_alpha = 1.2
    W_0_lbm = 1684.5
    W_pi_lbm = 17.7
    W_alpha_lbm = 1662.2

    W_bare_lbm = (
                         m_dot_core_corrected_lbm_per_sec / 100
                 ) ** b_m * (
                         W_0_lbm +
                         W_pi_lbm * (overall_pressure_ratio / 30) ** b_pi +
                         W_alpha_lbm * (bypass_ratio / 5) ** b_alpha
                 )
    W_bare = W_bare_lbm / kg_to_lbm

    ##### Compute nacelle weight

    ### Nondimensional parameters, given by Drela in TASOPT doc.
    r_s_nace = 12
    f_inlet = 0.4
    f_fan = 0.2
    f_exit = 0.4
    r_core = 12

    ### Fan size in dumb units
    d_fan_ft = diameter_fan * m_to_ft
    d_fan_in = d_fan_ft * 12

    ### Compute the diameter of the LPC based on fan diameter and BPR.
    d_LPC_ft = d_fan_ft * (bypass_ratio) ** -0.5

    ### Models from Drela in TASOPT
    S_nace_sqft = r_s_nace * np.pi * (d_fan_ft / 2) ** 2

    A_inlet_sqft = f_inlet * S_nace_sqft
    A_fan_sqft = f_fan * S_nace_sqft
    A_exit_sqft = f_exit * S_nace_sqft
    A_core_sqft = r_core * np.pi * (d_LPC_ft / 2) ** 2

    W_inlet_lbm = A_inlet_sqft * (2.5 + 0.0238 * d_fan_in)
    W_fan_lbm = A_fan_sqft * 1.9
    W_exit_lbm = A_exit_sqft * (2.5 * 0.0363 * d_fan_in)
    W_core_lbm = A_core_sqft * 1.9

    W_nace_lbm = W_inlet_lbm + W_fan_lbm + W_exit_lbm + W_core_lbm
    W_nace = W_nace_lbm / kg_to_lbm

    ##### Compute accessory and pylon weights

    ### Nondimensional parameters, given by Drela in TASOPT doc
    f_add = 0.10
    f_pylon = 0.10

    W_add = f_add * W_bare
    W_pylon = f_pylon * (W_bare + W_add + W_nace)

    ##### Compute the total weight
    W_engine = W_bare + W_add + W_nace + W_pylon

    return W_engine


def m_dot_corrected_over_m_dot(
        temperature_total_2,
        pressure_total_2,
):
    """
    Computes the ratio `m_dot_corrected / m_dot`, where:

        * `m_dot_corrected` is the corrected mass flow rate, where corrected refers to correction to ISO 3977 standard
        temperature and pressure conditions (15C, 101325 Pa).

        * `m_dot` is the raw mass flow rate, at some other conditions.

    Args:

        temperature_total_2: The total temperature at the compressor inlet face, at the conditions to be evaluated. [K]

        pressure_total_2: The total pressure at the compressor inlet face, at the conditions to be evaluated. [Pa]

    Returns:

        The ratio `m_dot_corrected / m_dot`.

    """
    temperature_standard = 273.15 + 15
    pressure_standard = 101325
    return (
            temperature_total_2 / temperature_standard
    ) ** 0.5 / (pressure_total_2 / pressure_standard)


if __name__ == '__main__':
    import aerosandbox as asb

    atmo = asb.Atmosphere(altitude=10668)
    op_point = asb.OperatingPoint(atmo, velocity=0.80 * atmo.speed_of_sound())
    m_dot_corrected_over_m_dot_ratio = m_dot_corrected_over_m_dot(
        temperature_total_2=op_point.total_temperature(),
        pressure_total_2=op_point.total_pressure()
    )

    ### CFM56-2 engine test
    mass_cfm56_2 = mass_turbofan(  # Data here from Wikipedia, cross-referenced to other sources for sanity check.
        m_dot_core_corrected=364 / (5.95 + 1),
        overall_pressure_ratio=31.2,
        bypass_ratio=5.95,
        diameter_fan=1.73
    )  # real mass: (2139 to 2200 kg bare, ~3400 kg installed)



import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from typing import Tuple, Dict, Union, Callable, List
from scipy import interpolate
from aerosandbox.tools.string_formatting import eng_string
import seaborn as sns


def pie(
        values: Union[np.ndarray, List[float]],
        names: List[str],
        colors: Union[np.ndarray, List[str]] = None,
        label_format: Callable[[str, float, float], str] = lambda name, value, percentage: name,
        sort_by: Union[np.ndarray, List[float], str, None] = None,
        startangle: float = 0.,
        center_text: str = None,
        x_labels: float = 1.25,
        y_max_labels: float = 1.3,
        arm_length=20,
        arm_radius=5,
):
    # TODO docs
    ax = plt.gca()
    n_wedges = len(values)


    ### Check inputs
    if not len(names) == n_wedges:
        raise ValueError()  # TODO

    ### Sort by
    sort_by_error= ValueError('''Argument `sort_by` must be one of:\n
    * a string of "values", "names", or "colors"
    * an array of numbers corresponding to each pie slice, which will then be used for sorting
    ''')

    if sort_by is None:
        sort_by = np.arange(n_wedges)
    elif sort_by == "values":
        sort_by=values
    elif sort_by=="names":
        sort_by=names
    elif sort_by=="colors":
        sort_by=colors # this might not make sense, depending on
    elif isinstance(sort_by, str):
        raise sort_by_error

    order = np.argsort(sort_by)
    names = np.array(names)[order]
    values=np.array(values)[order]

    if colors is None:
        # Set default colors
        colors = sns.color_palette("husl", n_colors=n_wedges)
    else:
        colors=np.array(colors)[order]

    ### Compute percentages
    values = np.array(values).astype(float)
    total = np.sum(values)
    percentages = 100 * values / total

    wedges, texts = ax.pie(
        x=values,
        colors=colors,
        startangle=startangle,
        wedgeprops=dict(
            width=0.3
        )
    )

    for w in wedges:
        w.theta_mid = (w.theta1 + w.theta2) / 2
        w.x_pie = np.cos(np.deg2rad(w.theta_mid))
        w.y_pie = np.sin(np.deg2rad(w.theta_mid))
        w.is_right = w.x_pie > 0

    left_wedges = [w for w in wedges if not w.is_right]
    right_wedges = [w for w in wedges if w.is_right]

    y_texts_left = y_max_labels * np.linspace(-1, 1, len(left_wedges))
    y_texts_right = y_max_labels * np.linspace(-1, 1, len(right_wedges))

    if len(left_wedges) == 1:
        y_texts_left = [w.y_pie for w in left_wedges]
    if len(right_wedges) == 1:
        y_texts_right = [w.y_pie for w in right_wedges]

    left_wedge_order = np.argsort([w.y_pie for w in left_wedges])
    for i, w in enumerate(np.array(left_wedges, "O")[left_wedge_order]):
        w.y_text = y_texts_left[i]

    right_wedge_order = np.argsort([w.y_pie for w in right_wedges])
    for i, w in enumerate(np.array(right_wedges, "O")[right_wedge_order]):
        w.y_text = y_texts_right[i]

    for i, w in enumerate(wedges):
        x_text = x_labels * np.sign(w.x_pie)
        ax.annotate(
            text=label_format(names[i], values[i], percentages[i]),
            xy=(w.x_pie, w.y_pie),
            xytext=(x_text, w.y_text),
            horizontalalignment="left" if w.is_right else "right",
            arrowprops=dict(
                arrowstyle="-",
                color="k",
                connectionstyle=f"arc,angleA={180 if w.is_right else 0},angleB={w.theta_mid},armA={arm_length},armB={arm_length},rad={arm_radius}",
                relpos=(
                    0 if w.is_right else 1,
                    0.5
                )
            ),
            va="center",
        )

    if center_text is not None:
        plt.text(
            x=0,
            y=0,
            s=center_text,
            ha="center",
            va="center",
            fontsize=16,
        )


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    data = {
        "USA"               : 336997624,
        "Mexico"            : 126705138,
        "Canada"            : 38115012,
        "Guatemala"         : 17608483,
        "Haiti"             : 11447569,
        "Cuba"              : 11256372,
        "Dominican Republic": 11117873,
        "Honduras"          : 10278345,
        "Nicaragua"         : 6850540,
        "El Salvador"       : 6314167,
        "Costa Rica"        : 5153957,
        "Panama"            : 4351267,
    }
    data["Other"] = 597678511 - np.sum(np.array(list(data.values())))

    fig, ax = plt.subplots(figsize=(9, 5))
    pie(
        values=list(data.values()),
        names=list(data.keys()),
        colors=[
            "navy" if s in ["USA"] else "lightgray"
            for s in data.keys()
        ],
        label_format=lambda name, value, percentage: f"{name}, {eng_string(value)} ({percentage:.0f}%)",
        startangle=40,
        center_text="Majority of North\nAmerica's Population\nlives in USA"
    )
    p.show_plot()

#     import pandas as pd
#     from io import StringIO
#
#     df = pd.read_csv(
#         StringIO("""\
# person,slices eaten,gender
# alice,9,woman
# bob,6,man
# charlie,5,man
# dan,8,man
# eve,7,woman
# frank,9,man
# grace,4,woman
# heidi,3,woman
# """)
#     )
#
#     fig, ax = plt.subplots(figsize=(8, 5))
#     pie(
#         values=df['slices eaten'],
#         names=df['person'],
#         colors=['blue' if g == 'man' else 'red' for g in df['gender']],
#         label_format=lambda n, v, p: f"{n.capitalize()}, {v:.0g} ({p:.0f}%)",
#         # sort_by=df[]
#     )
#     p.show_plot()
